#include <stdio.h>
#include "cylinder.h"

// Default size, resolution and color of an etCylinder.
const double DEFAULT_RADIUS = 0.025; // 50mm diameter.
const double DEFAULT_HEIGHT = 0.05; // 50mm.

// Returns the force to be applied.  See etForceRenderer.
etVector etCylinder::computePointForce(etVector l_position)
{
	
	if (!cull(l_position)) {
		double l_height = this->getHeightRaw();
		double l_radius = this->getRadiusRaw();
		
		etVector l_posXZ(l_position.x(), 0, l_position.z());
		
		//if l_radialDistance and l_yDistance are both >0, we are
		//inside the cylinder
		
		// Calculate the distance from the side.
		double l_radialDistance = l_radius - l_posXZ.length();
		double l_yDistance;
		
		// Calculate the distance from the top or bottom.
		if (l_position.y() > 0) {// Top
			l_yDistance = l_height * 0.5 - l_position.y();
		}
		else {// Bottom
			l_yDistance = l_height * 0.5 + l_position.y();
		}
		
		etVector l_force;
		
		// If inside the cylinder.
		if ( l_yDistance > 0 && l_radialDistance > 0) {
			// If closest to side.
			if (l_radialDistance < l_yDistance) {
				l_force = l_posXZ.normalize();
				l_force = l_force.scale(l_radialDistance);
				etVector temp = l_posXZ.normalize().scale(l_radius); 
				scp = etVector(temp.x(), l_position.y(), temp.z());
			}
			// If closest to top.
			else if (l_position.y() > 0) {
				l_force = etVector(0,l_yDistance,0);
				scp = etVector(l_position.x(), l_height * 0.5, l_position.z());
			}
			// If closest to bottom.
			else {
				l_force = etVector(0,-l_yDistance,0);
				scp = etVector(l_position.x(), -(l_height * 0.5), l_position.z());
			}

			//Calculate Friction	
			if (!doFriction) {
				// Set start values.
				setLastStylusPosition(l_position);
				setLastStickPoint(scp);
				doFriction = true;
			}
			else {
				// Calculate friction.
				etVector frictionForce(0,0,0);
				etVector distance = scp - getLastStickPoint();
				double scaleFactor = l_force.length() * 10;
				float threshold_one = 0.01;
				
				if (distance.length() < threshold_one) {
					// No friction for small movements.
				}
				else if (distance.length() < staticFriction) {
					// Static friction.
					distance = distance - distance.normalize().scale(threshold_one);
					frictionForce = distance.scale(-1.0);
					if (scaleFactor > 1)
						scaleFactor = 1;
					frictionForce = frictionForce.scale(scaleFactor);
					l_force += frictionForce;
				}
				else {
					// Dynamic friction.
					stickPoint = getLastStickPoint()+distance.scale(0.05);
					setLastStickPoint(stickPoint);
					frictionForce = distance.normalize().scale(-dynamicFriction);
					if (scaleFactor > 1)
						scaleFactor = 1;
					frictionForce = frictionForce.scale(scaleFactor);
					l_force += frictionForce;
				}
			}
			setLastSCP(scp);
		}
		
		l_force = l_force.scale(this->getStiffnessRaw() );
		
		return l_force;
	}
	else
		return etVector(0,0,0);
}

// Cull.
bool etCylinder::cull(etVector l_position)
{

    double l_height = this->getHeightRaw();
    double l_radius = this->getRadiusRaw();

    return (fabs(l_position.y()) > l_height*0.5 ||
            sqrt(l_position.x() * l_position.x() +
                 l_position.z() * l_position.z()) > l_radius);
}

// Constructor
etCylinder::etCylinder()
    : etObject(),
      m_Radius(DEFAULT_RADIUS),
      m_Height(DEFAULT_HEIGHT)
{
}

// Copy Constructor
etCylinder::etCylinder(const etCylinder& other)
    : etObject(other),
      m_Radius(other.getRadius()),
      m_Height(other.getHeight())
{
}

// Assignment operator.
etCylinder& etCylinder::operator=(const etCylinder& other)
{
    etObject::operator=(other);
    setRadius(other.getRadius());
    setHeight(other.getHeight());
    return (*this);
}

//Returns the tool associated with the graphic renderer.
etCylinder* etCylinder::clone() const
{
    return new etCylinder(*this);
}
