// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// edge_data.cpp -- contains the edge_data implementation
//
// $Revision: #2 $
// $Change: 1826 $
//

#include <et/common.h>
#include "orientation.h"

#include "triangle.h"
#include "edgedata.h"//<et/objects/edge_data.h>

// Cutoff for the angle between triangles which
// triggers force towards common triangle edge rather
// than interpolating.
const double POLYSURF_VERY_ACUTE_ANGLE_RADIANS = etOrientation::PII * 0.4;

int PolySurf::EdgeData::getNumNeighbours() const
{
    int l_result = 0;
    if (m_neighbour0 != 0)
    {
        l_result++;
    }
    if (m_neighbour1 != 0)
    {
        l_result++;
    }

    return l_result;
}

void PolySurf::EdgeData::addNeighbour(Triangle* a_newNeighbour)
{
    if (m_neighbour0 == 0)
    {
        m_neighbour0 = a_newNeighbour;
    }
    else if (m_neighbour1 == 0)
    {
        m_neighbour1 = a_newNeighbour;
    }
}
void PolySurf::EdgeData::init(const Vertex* a_vertex0, 
                                      const Vertex* a_vertex1, 
                                      Triangle* a_neighbour0, 
                                      Triangle* a_neighbour1)
{
    etVector l_point0 = a_vertex0->m_position;
    etVector l_point1 = a_vertex1->m_position;

    m_neighbour0 = a_neighbour0;
    m_neighbour1 = a_neighbour1;

    etVector l_normal0;
    etVector l_normal1; 
    
    Triangle* l_neighbour = 0;
    bool l_areCheckingNeighbour0 = true;

    if (a_neighbour0 != 0)
    {
        l_normal0 = m_neighbour0->getNormal();        
        l_neighbour = m_neighbour0;
    }
    if (a_neighbour1 != 0)
    {
        l_normal1 = m_neighbour1->getNormal();           
        
        if (l_neighbour == 0)
        {
            l_neighbour = m_neighbour1;
            l_areCheckingNeighbour0 = false;
        }
    }
    
    bool l_isEdgeFull = (getNumNeighbours() == 2);

    // If the triangles have the same vertex ordering (CW vs CCW)
    // the reverse one of the normals
    if (l_isEdgeFull && 
        (isTriangleOrdered(a_neighbour0, a_vertex0, a_vertex1) ==
        isTriangleOrdered(a_neighbour1, a_vertex0, a_vertex1)) )
    {    
        l_normal1 = l_normal1.scale(-1);
        m_areNeighbourNormalsFlipped = true;
    }

    // Create perp edge planes for each neighbour
    if (m_neighbour0 != 0)
    {
        if (isTriangleOrdered(a_neighbour0, a_vertex0, a_vertex1))
        {
            m_perpEdgePlane0 = Plane(
                    l_point0,
                    l_point1,
                    l_point0 + m_neighbour0->getNormal());
        }
        else
        {
            m_perpEdgePlane0 = Plane(
                    l_point1,
                    l_point0,
                    l_point0 + m_neighbour0->getNormal());
        }
    }
    if (m_neighbour1 != 0)
    {
        if (isTriangleOrdered(a_neighbour1, a_vertex1, a_vertex0))
        {
            m_perpEdgePlane1 = Plane(
                    l_point1,
                    l_point0,
                    l_point1 + m_neighbour1->getNormal());
        }
        else
        {
            m_perpEdgePlane1 = Plane(
                    l_point0,
                    l_point1,
                    l_point0 + m_neighbour1->getNormal());
        }
    }

    // Create the edge planes from three points,
    // the two points of the edge and point0 + average of
    // the neigbouring triangle normals
    if (getNumNeighbours() > 0)
    {             
        etVector l_point2 = l_point0 + l_normal0 + l_normal1;
        m_edgePlane = Plane(l_point0, l_point1, l_point2);
            
        // Replace the edgeplane with a reversed one
        // if the normal  doesn't point into neighbour1
        for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
        {
            // Get the vertex from neighbour which isn't in the edge
            const Vertex* l_oppositeVert = l_neighbour->getVertex(l_cornerIndex);
            if (l_oppositeVert != a_vertex0 &&
                l_oppositeVert != a_vertex1)
            {
                double l_distance = getEdgePlane().distance(l_oppositeVert->m_position);
                if (( l_distance > 0 && l_areCheckingNeighbour0) ||
                    ( l_distance < 0 && !l_areCheckingNeighbour0))
                {
                    m_edgePlane = Plane(l_point1, l_point0, l_point2);
                }
                break;
            }
        }     
    }

    if (l_isEdgeFull)
    {
        // Test for acute angle between the neighbourig triangles being < 90 degrees            
        //
        // Assumes that the normals of the neighbours are corretly 
        // aligned.
        
        double l_dot = l_normal0.dot(l_normal1);
        if (l_dot <= 0)
        // Normal angle > 90deg, hence the plane angle < 90deg
        {
            // Set edge as acute (ie small angle between neighbours is < 90)
            m_isAcute = true;
            
            double l_angle = etOrientation::PII - acos(l_dot);           
            if (l_angle < POLYSURF_VERY_ACUTE_ANGLE_RADIANS)
            {
                m_isVeryAcute = true;
            }
        }
                
        // Get the points from neighbours which aren't in this edge
        // XXX repeated in getAngle
        etVector l_otherPoint0;
        etVector l_otherPoint1;
        for(int l_cornerIndex=0; l_cornerIndex< 3; l_cornerIndex++)
        {
            if ((m_neighbour0->getVertex(l_cornerIndex) != a_vertex0) &&
                (m_neighbour0->getVertex(l_cornerIndex) != a_vertex1))
            {
                l_otherPoint0 = m_neighbour0->getVertexPosition(l_cornerIndex);
            }
            if ((m_neighbour1->getVertex(l_cornerIndex) != a_vertex0) &&
                (m_neighbour1->getVertex(l_cornerIndex) != a_vertex1))
            {
                l_otherPoint1 = m_neighbour1->getVertexPosition(l_cornerIndex);
            }
        }

        // Set the edge to be convex if the
        // neighbour0's normal points away from neighbour1.
        if ((l_otherPoint1 - l_otherPoint0).dot(l_normal0) < 0)
        {
            // Then normals point outwards, 
            // hence the edge is convex
            m_isConvex = true;
        }        
    }
}    
    
