// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// edge_data.h -- contains the EdgeData class.
//
// $Revision: #1 $
// $Change: 1787 $
//

#ifndef POLYSURF_EDGE_DATA_H
#define POLYSURF_EDGE_DATA_H

#include <et/objects/plane.h>

namespace PolySurf
{
    // Forward declarations
    struct Vertex;
    class Triangle;

    //
    // Class:
    //     EdgeData
    //
    // Description:
    //     EdgeData records the details of a join between
    //     two adjacent triangles.
    //
    // Exceptions:
    //     None
    //
    // Notes:
    //     None
    //
    // See Also Classes:
    //     Triangle, etHapticTrisurface
    //
    // See Also Functions:
    //     None
    //
    class EdgeData
    {           
        public:              

            // Constructor
            EdgeData(): 
                m_neighbour0(0),
                m_neighbour1(0),
                m_isAcute(false),
                m_isVeryAcute(false),
                m_isConvex(false),
                m_areNeighbourNormalsFlipped(false)
            {
                // do nothing
            }

            // Initialise the edge data,             
            void init(const Vertex* a_vertex0,
                      const Vertex* a_vertex1,
                      Triangle* neighbour0,
                      Triangle* neighbour1);

            // Return the number of neighbouring triangles.
            int getNumNeighbours() const;
        
            // Add a new neighbouring triangle.
            void addNeighbour(Triangle* a_newNeighbour);
       
            // Set the edge planes
            void setEdgePlane(Plane a_edgePlane)
            {
                m_edgePlane = a_edgePlane;
            }

            // Return the edge plane, which will be a plane composed
            // from the end points of the edge and the average
            // of the two neighbours normals.
            inline Plane getEdgePlane() const
            {
                return m_edgePlane;
            }

            // Returns a plane through the edge which is
            // perpindicular to neighbour0.
            inline Plane getPerpEdgePlane0() const
            {
                return m_perpEdgePlane0;
            }
             
            // Returns a plane through the edge which is
            // perpindicular to neighbour1.
            inline Plane getPerpEdgePlane1() const
            {
                return m_perpEdgePlane1;
            }

            // Returns neighbour0
            inline Triangle* getNeighbour0() const
            {
                return m_neighbour0;
            }

            // Returns neighbour1
            inline Triangle* getNeighbour1() const
            {
                return m_neighbour1;
            }
                    
            // Returns the neighbour which isn't a_triangle
            inline Triangle* getOtherNeighbour(Triangle* a_triangle) const
            {
                Triangle* l_result;
                if (a_triangle == m_neighbour0)
                {
                    l_result = m_neighbour1;
                }
                else
                {
                    l_result = m_neighbour0;
                } 

                return l_result;
            }

            // Returns true if the smaller angle between the triangles
            // is less then 0.5 PI radians.
            inline bool isAcute() const{return m_isAcute;}

            // Returns true if the smaller angle between the triangles
            // is less then 0.4 PI radians.
            inline bool isVeryAcute() const{return m_isVeryAcute;}
            
            // Returns true if neighbour0's normals points into
            // the smaller angle between the triangle.
            inline bool isConvex() const{return m_isConvex;}

            // Return true if the vertex ordering is reversed between the neighbours.
            // (Clockwise vs Counter-Clokwise).
            inline bool areNeighbourNormalsFlipped() const{return m_areNeighbourNormalsFlipped;}

            // Returns true if the tool is touching the large angled side
            // of the edge.
            inline bool isContactOutsideEdge(Triangle* m_currentTriangle, bool a_isFrontFaceTouched) const
            {
                bool l_result = (isConvex() == a_isFrontFaceTouched);
                if (areNeighbourNormalsFlipped() && getNeighbour1() == m_currentTriangle)
                {
                    l_result = !l_result;
                }
            
                return l_result;                    
            }

        private:
            // Edge plane normal points away from neighbour0
            Plane m_edgePlane;
    
            // The angle between the neighbours is < 90 degrees
            bool m_isAcute;

            // The angle between the neighbours is much less then 90 degrees
            bool m_isVeryAcute;

            // If the edge is convex the neighbors normals
            // point away from the acute angled side.
            // This is only set if m_isAcute is true.
            bool m_isConvex;

            // True if the shared vertices appear in the same
            // order in each neighbour, which implies
            // that the normals are flipped between neighbours
            bool m_areNeighbourNormalsFlipped;

            // Planes perpindiuclar to the neighbours. facing outwards
            Plane m_perpEdgePlane0;
            Plane m_perpEdgePlane1;

            // Doesn't record the side
            Triangle* m_neighbour0;
            Triangle* m_neighbour1;
    };
}

#endif
