//=============================================================================
//   Filename : gstScene.h
//    Project : GHOST library v3.0
//    Authors : Chris Tarr
//    Revised : W. A. Aviles
// ----------------------------------------------------------------------------
// Description: Manages dynamic/haptic simulation and owns haptic
//              process or thread.  Also, arbitrates communication to
//              and from the haptic process or thread.
//
// Send your questions, comments or bugs to:
//    support@sensable.com
//    http://www.sensable.com
//
// Copyright (c) 1996-1999 SensAble Technologies, Inc. All rights reserved.
//=============================================================================

#ifndef GST_SCENE
#define GST_SCENE

#include <gstBasic.h>
#include <gstTransform.h>

// Holds haptic scene and mediates interaction with the PHANToM and servo loop.
class  GHOST_DLL_IMPORT_EXPORT gstScene {

public:

	enum {
		SMOOTHING_OFF,
		SMOOTHING_ON
	};

	// Constructor.
	gstScene();

	// Performs a recursive delete on all nodes under the root
	~gstScene();

    // lock() a gstScene before making changes to it if the servo loop
    // is running.  NOTE: Changes to the phantom or its ancestor nodes are
    // not supported when the servo loop is running, even if the scene is
    // locked.  Also, effects and manipulators will still be run even
    // if the scene is locked, so if this is incompatible with the
    // changes being made they should be stopped before making changes.
    // Finally, force fields are not enabled when the scene is locked.
    // returns TRUE if lock acquired, FALSE otherwise (if servo loop stopped)
	gstBoolean lock();

	// Call unlock() after a lock() to resume normal servo loop behavior.
    // NOTE: the scene should remain locked for only a few servo loop
    // iterations (each iteration is 1ms) to maintain an accurate haptic
    // simulation.
	void unlock();

	// Set root node of haptic scene graph.
	void setRoot(gstTransform *newRoot);

	// Get root node of haptic scene graph.
	gstTransform *getRoot() {
		return rootNode;
	}

	// FOR_GHOST_INTERNAL:
	// Turns ON forces of all PHANToMs in the scene graph.
	void turnForcesOn() {
		forcesOn = TRUE;
		firstLoop = TRUE;
	}

	// FOR_GHOST_INTERNAL:
	// Turns OFF forces of all PHANToMs in the scene graph.
	void turnForcesOff() {
		forcesOn = FALSE;
		firstLoop = TRUE;
	}

	// Set smoothing level for scene.  Smoothing attempts to 
	// remove high-frequency variations in gstPolyMesh geometries.
	// Currently only 0 (OFF) and 1 (ON) are supported.
	void setSmoothing(int newLevel) {
		smoothingLevel = newLevel;
	}

	// Get smoothing level of the scene.
	int getSmoothing() const {
		return smoothingLevel;
	}

	// Has simulation end/continue from occurence of a device
	// fault from any gstPHANToM in scene graph.  If flag is TRUE
	// then simulation will exit if any gstPHANToM in the scene
	// graph has a device fault.  If flag is FALSE, then device
	// faults from any gstPHANToM in the scene graph will cause
	// all forces to be turned off, but simulation will be
	// maintained.  As soon as the all gstPHANToMs in the scene
	// graph have no device faults, forces will be restarted.
	// If a device fault occurs will trying to start forces,
	// simulation will exit regardless of the value of flag.
	// Note: Device faults occur when the PHANToM device is
	// not hooked up properly to the computer, or when the remote
	// switch is not activated.
    void setQuitOnDevFault(gstBoolean flag) {
		quitOnDevFault = flag;
	}

	// Returns the current value of quitOnDevFault.
	gstBoolean getQuitOnDevFault() const {
		return quitOnDevFault;
	}

    // Calls the graphics callbacks for all nodes in the scene
    // which have changed since the last call to updateGraphics().
    // This means nodes which have not changed will not have their
    // graphics callback called.
	void updateGraphics();

    // Calls the event callback for all nodes in the scene
    // which have new events since the last call to updateEvents().
    // Each callback is called once for each new event.
	void updateEvents();
         
	// Start the haptic simulation as separate process.
	// Control is returned immediately.
	int startServoLoop();

	// Clean up the haptic simulation after an internal error.
	void cleanupServoLoop();

	// Stop the haptic simulation.
	void stopServoLoop();


	// FOR_GHOST_INTERNAL:
	// Returns whether the haptics process is currently in
	//  the servoLoop.
	int getDoneOneLoop() {return doneOneLoop;}

	// Returns TRUE if haptics process has finished (i.e.
        // if the servo loop is not running).
	int getDoneServoLoop() const { return doneServoLoop;}

	  // Turns safety on and off.  If safety is off, the servo loop is
	  // allowed to take as much CPU time as necessary to finish.
	  // This may result in instability or crashes.  The default is TRUE (on).
	void setSafety(gstBoolean _s) {
		scene_safety = _s;
	}
  
	// Returns TRUE if safety limits are on.
	gstBoolean getSafety() {
		return scene_safety;
	}

gstInternal public:

	// FOR_GHOST_INTERNAL:
	// Used by system or for creating sub-classes only.
	// Servo loop body.
	int servoLoop();

	// FOR_GHOST_INTERNAL:
	// Used by system or for creating sub-classes only.
	void setDoneOneLoop(gstBoolean newVal) {doneOneLoop = newVal;}

protected:

	gstTransform			*rootNode;
	int						preparingGraphics, preparingEvents;
	int						doneServoLoop, doneOneLoop;
	int						dynamicF;
	double					dynamicFrictionAvg, staticFrictionAvg, surfaceDampingAvg;
	double					Kfriction;
	int						firstLoop, forcesOn;
	gstBoolean				needLock, locked;
	gstBoolean				quitOnDevFault;
	int						smoothingLevel;
    gstBoolean scene_safety;
};



#endif // GST_SCENE
