#pragma warning(disable:4786)
#pragma warning(disable:4503)

#include <et/common.h>
#include <et/time.h>


#include <utility>
#include <set>


#include "octtree.h"//<et/objects/oct_tree.h>
#include "triangle.h"//<et/objects/triangle.h>
#include "edgedata.h"//<et/objects/edge_data.h>
#include <et/objects/plane.h>
#include "matrix.h"//<et/matrix.h>


#include "liquidtrisurface.h"

const double POLYSURF_EPSILON = 0.00000001;
const double POLYSURF_INTERP_DISTANCE = 0.2;
const double POLYSURF_STIFFNESS = 500;
const bool POLYSURF_USE_OCTREE = true;
const int POLYSURF_NEIGHBOUR_CROSS_LIMIT = 11;
const double DEFAULT_VISCOSITY = 200;

// Ratio of vertex merge distance to average triangle side length
const double POLYSURF_VERTEX_MERGE_DISTANCE_RATIO = 0.01;

namespace{
	etVector closestPointToLine(const std::pair<etVector, etVector>& a_line, const etVector& a_point)
	{
		assert(a_line.second != a_line.first);
		etVector l_diff = a_line.second - a_line.first;
		
		return a_line.first +
			l_diff.scale((a_point - a_line.first).dot(l_diff) / l_diff.dot(l_diff));
	}
	
	// Assumes that point lies on the line
	bool isPointInsideLine(const std::pair<etVector, etVector>& a_line, const etVector& a_point)
	{        
		double l_edgeLength = (a_line.first - a_line.second).length();        
		if ((a_line.first - a_point).length() > l_edgeLength || 
			(a_line.second - a_point).length() > l_edgeLength)
		{
			return false;
		}
		
		return true;
	}
	
	PolySurf::Plane triangleToPlane(PolySurf::Triangle* a_triangle)
	{
		return PolySurf::Plane(a_triangle->getNormal(), a_triangle->getVertexPosition(0));
	}
	
	// The triangle test returns true if the new triangle 
	// is touched on the same face indicated by a_wasFrontFaceTouched
	class ReintersectingTriangleTest : public etOctTree::TriangleTest
	{
	public:
		
		ReintersectingTriangleTest(bool a_wasFrontFaceTouched, etVector a_toolPosition)
			:TriangleTest(),
			m_toolPosition(a_toolPosition),
			m_wasFrontFaceTouched(a_wasFrontFaceTouched)
		{
			// do nothing
		}
		
		virtual operator()(PolySurf::Triangle* a_triangle)
		{
			bool l_isFrontFaceTouched = 
				(PolySurf::Plane(a_triangle->getNormal(), a_triangle->getVertexPosition(0))
				.distance(m_toolPosition) < 0);
			
			// Return true if the new triangle is touched on the same
			// side as the old triangle
			return (l_isFrontFaceTouched == m_wasFrontFaceTouched);
		}
		
	private:
		bool m_wasFrontFaceTouched;
		etVector m_toolPosition;
	};       
	
	class Edge
	{   
		friend etLiquidTrisurface;
		
	public:
		inline const PolySurf::Vertex* getVertex0() const 
		{
			return m_v0;
		}
		inline const PolySurf::Vertex* getVertex1() const 
		{
			return m_v1;
		}
		
	protected:           
		explicit Edge(PolySurf::Vertex* a_v0, PolySurf::Vertex* a_v1)
			: m_v0(a_v0), 
			m_v1(a_v1)
		{
			if (m_v0 > m_v1)
			{
				PolySurf::Vertex* tmp = m_v0;
				m_v0 = m_v1;
				m_v1 = tmp;
			}
		}
		
	private:               
		PolySurf::Vertex* m_v0; // Lower index is always first
		PolySurf::Vertex* m_v1;             
	};
	
	inline bool operator<(const Edge& a, const Edge& b)
	{
		return a.getVertex0() < b.getVertex0() || 
			(a.getVertex0() == b.getVertex0() && 
			a.getVertex1() < b.getVertex1());
	}
	
	void mergeVertices(std::vector<PolySurf::Triangle*> a_triangles)
	{        
		std::map<PolySurf::Vertex*, PolySurf::Vertex*> l_repeatMap;
		
		std::set<PolySurf::Vertex*> l_vertices;
		
		// Get all vertices used in the set of triangles
		int l_triangleIndex;
		double l_totalSideLength = 0;
		
		for(l_triangleIndex = 0; l_triangleIndex < a_triangles.size(); l_triangleIndex++)
		{
			PolySurf::Vertex* l_point0 = a_triangles[l_triangleIndex]->getVertex(0);
			PolySurf::Vertex* l_point1 = a_triangles[l_triangleIndex]->getVertex(1);
			PolySurf::Vertex* l_point2 = a_triangles[l_triangleIndex]->getVertex(2);
			l_vertices.insert(l_point0);
			l_vertices.insert(l_point1);
			l_vertices.insert(l_point2);
			
			l_totalSideLength += (l_point0->m_position - l_point1->m_position).length();
			l_totalSideLength += (l_point1->m_position - l_point2->m_position).length();
			l_totalSideLength += (l_point2->m_position - l_point0->m_position).length();
		}
		
		double l_mergeDistance = l_totalSideLength / (a_triangles.size()) / 3;
		l_mergeDistance *= POLYSURF_VERTEX_MERGE_DISTANCE_RATIO;
		
		// Square it 
		l_mergeDistance *= l_mergeDistance;
		
		//find all repeated vert positions
		std::set<PolySurf::Vertex*>::iterator l_outerIter = l_vertices.begin();
		std::set<PolySurf::Vertex*>::iterator l_innerIter;
		
		while(l_outerIter != l_vertices.end())
		{
			l_innerIter = l_outerIter;
			l_innerIter++;
			while(l_innerIter != l_vertices.end())
			{
				etVector l_diff = (*l_innerIter)->m_position - 
					(*l_outerIter)->m_position;
				
				// If locations match                 
				if (l_diff.dot(l_diff) < l_mergeDistance)
				{
					l_repeatMap[(*l_outerIter)] = (*l_innerIter);
				}
				l_innerIter++;
			}
			l_outerIter++;
		}
		
		for(l_triangleIndex = 0; l_triangleIndex < a_triangles.size(); l_triangleIndex++)
		{
			for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
			{
				std::map<PolySurf::Vertex*, PolySurf::Vertex*>::iterator l_iter = 
					l_repeatMap.find( a_triangles[l_triangleIndex]->getVertex(l_cornerIndex) );
				
				if (l_iter != l_repeatMap.end())
				{
					a_triangles[l_triangleIndex]->setVertex(l_cornerIndex, (*l_iter).second);
				}               
			}                
		}
		//XXX should remove all repeats from the vertex list in the surface
	}
	
	class MergeVerticesInNode : public etOctTree::ItemListOperation
	{
		virtual operator()(etOctTree::item_list_type& a_itemList)
		{
			mergeVertices(a_itemList);
		}
	};
}
//
// etHapticTrisurface Member functions
//

etLiquidTrisurface::etLiquidTrisurface()
: etObject(),         
m_isFrontFaceTouchEnabled(true),
m_isBackFaceTouchEnabled(true),    
m_ToolContactPosition(etVector()),     
m_Scale(1),
m_octTree(0),
m_CurrentTriangle(0),    
m_isDrawCurrentTriangleEnabled(true),
m_Stiffness(POLYSURF_STIFFNESS),
m_prevPosition(etVector()),
m_isForceToVertActive(false),      
m_currentVertex(-1),
m_currentTriangle(0),
m_isFirstTime(true),
viscosity(DEFAULT_VISCOSITY)
{
}

etLiquidTrisurface::etLiquidTrisurface(const etLiquidTrisurface& other)
:etObject(other),          
m_ToolContactPosition(etVector()),    
m_Scale(other.getScale()),
m_CurrentTriangle(0),        
m_Stiffness(POLYSURF_STIFFNESS)
{ 
}

etLiquidTrisurface& etLiquidTrisurface::operator=(const etLiquidTrisurface& other)
{   
	//XXX finish 
	return (*this);
}

void etLiquidTrisurface::clearTrianglesAndVertices()
{
	int l_index;
	for(l_index = 0; l_index < m_triangles.size(); l_index++)
	{
		delete m_triangles[l_index];        
	}
	m_triangles.clear();
	
	for(l_index = 0; l_index < m_vertices.size(); l_index++)
	{
		delete m_vertices[l_index];        
	}
	m_vertices.clear();
}

etLiquidTrisurface::~etLiquidTrisurface()
{
	// Delete all edge data
	clearEdgeData();
	
	// Delete all triangles and vertices
	clearTrianglesAndVertices();
	
	// Delete the octtree
	delete m_octTree;
}

void etLiquidTrisurface::setVertices(const std::vector<etVector>& a_vertices)
{
	m_vertices.clear();
	
	int l_index;
	for(l_index = 0; l_index < a_vertices.size(); l_index++)
	{
		m_vertices.push_back(new PolySurf::Vertex(a_vertices[l_index]));
	}
}

void etLiquidTrisurface::create(std::vector<etVector> a_vertices,
								std::vector<int> a_coordinates,
								bool a_isVertexMergeEnabled)
{
	// Clear the current triangle land vertex lists
	clearTrianglesAndVertices();
	
	int l_index;
	for(l_index = 0; l_index < a_vertices.size(); l_index++)
	{
		m_vertices.push_back(new PolySurf::Vertex(a_vertices[l_index]));
	}
	
	int l_cornerIndex = 0;
	PolySurf::Vertex* l_vertices[3];
	for(l_index = 0; l_index < a_coordinates.size(); l_index++)
	{
		// Record indices until -1 read or
		// three indices found
		
		if (l_cornerIndex < 3)
		{
			l_vertices[l_cornerIndex] = m_vertices[a_coordinates[l_index]];
			l_cornerIndex++;
		}
		// ignore indices until the next -1
		else if (a_coordinates[l_index] == -1)
		{
			l_cornerIndex = 0;
			addTriangle(l_vertices[0], l_vertices[1], l_vertices[2]);               
		}     
	}
	
	init(a_isVertexMergeEnabled);
}

// Adds a triangle using indices into the vertex list
void etLiquidTrisurface::addTriangle(int a_index0, int a_index1, int a_index2)
{
	// Add the triangle if all indices are valid
	if (a_index0 > -1 && a_index0 < m_vertices.size() && 
		a_index1 > -1 && a_index1 < m_vertices.size() && 
		a_index2 > -1 && a_index2 < m_vertices.size() )
	{
		addTriangle(
			m_vertices[a_index0],
			m_vertices[a_index1],
			m_vertices[a_index2]);
	}
}

// Adds a triangle, assumes the Vertices are already handled
// Records edge neighbours
void etLiquidTrisurface::addTriangle(PolySurf::Vertex* a_point0, PolySurf::Vertex* a_point1, PolySurf::Vertex* a_point2)
{
	PolySurf::Triangle* l_triangle = new PolySurf::Triangle(a_point0, a_point1, a_point2);
	
	m_triangles.push_back(l_triangle);    
}

void etLiquidTrisurface::initOctTree()
{
	// Init the octree
	if (POLYSURF_USE_OCTREE)
	{
		if (m_octTree)
		{
			delete m_octTree;
		}
		m_octTree = new etOctTree();
		m_octTree->addTriangles(m_triangles);
	}
}

void etLiquidTrisurface::clearEdgeData()
{
	// Remove edge data refs from all triangles
	for(int l_triangleIndex = 0; l_triangleIndex < m_triangles.size(); l_triangleIndex++)
	{
		for(int l_edgeIndex = 0; l_edgeIndex < 6; l_edgeIndex++)
		{
			m_triangles[l_triangleIndex]->setEdgeData(l_edgeIndex, 0);          
		}
	}
	
	// Delete all edgeDatas
	for(int l_index = 0; l_index < m_edgeDataList.size(); l_index++)
	{           
		delete m_edgeDataList[l_index];
	}
	
	// Clear the list
	m_edgeDataList.clear();
}

// Calculate edge planes for all edges
void etLiquidTrisurface::processEdges()
{           
	// For each edge
	//      Set the edge plane using the two vertices
	//      and the average of the normals from the two 
	//      neighbouring triangles.    
	//      
	//      Record the angle between the edges   
	//      Single neighbour edges have an orthogonal edge plane
	
	// Clear current edge datas
	clearEdgeData();
	
	// Map each Edge to a list of triangles and index wihtin the triangle
	typedef std::map<Edge, std::vector<std::pair<PolySurf::Triangle*, int> > > edge_map_type;
	edge_map_type l_edgeList;   
	
	// Add the edges of all current triangles to the edge map
	for(int l_triangleIndex = 0; l_triangleIndex < m_triangles.size(); l_triangleIndex++ )
	{
		PolySurf::Triangle* l_tri = m_triangles[l_triangleIndex];
		Edge l_triangleEdges[] ={
			Edge(l_tri->getVertex(0), l_tri->getVertex(1)),
				Edge(l_tri->getVertex(1), l_tri->getVertex(2)),
				Edge(l_tri->getVertex(2), l_tri->getVertex(0))};
			
			for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
			{           
				std::pair<PolySurf::Triangle*, int> l_edgeLocation(l_tri, l_cornerIndex);
				l_edgeList[l_triangleEdges[l_cornerIndex]].push_back(l_edgeLocation);
			}
	}
	
	edge_map_type::iterator l_iter = l_edgeList.begin();    
	while(l_iter != l_edgeList.end())
	{
		const PolySurf::Vertex* l_vertex0 = (*l_iter).first.getVertex0();
		const PolySurf::Vertex* l_vertex1 = (*l_iter).first.getVertex1(); 
		
		//if edge has < 3 neighbours        
		if ((*l_iter).second.size() < 3)
		{
			PolySurf::Triangle* l_neighbour0 = (*l_iter).second[0].first;
			PolySurf::Triangle* l_neighbour1 = 0;
			if ((*l_iter).second.size() > 1)
			{
				l_neighbour1 = (*l_iter).second[1].first;
			}
			
			PolySurf::EdgeData* l_edgeData = new PolySurf::EdgeData;
			m_edgeDataList.push_back(l_edgeData);
			l_edgeData->init(l_vertex0,
				l_vertex1, 
				l_neighbour0, 
				l_neighbour1);
			
			//add to triangle
			int l_edgeIndex = (*l_iter).second[0].second;
			
			l_neighbour0->setEdgeData(l_edgeIndex, l_edgeData);
			l_neighbour0->setEdgeData(l_edgeIndex+3, l_edgeData);
			
			if ((*l_iter).second.size() > 1)
			{
				l_edgeIndex = (*l_iter).second[1].second;
				
				l_neighbour1->setEdgeData(l_edgeIndex, l_edgeData);
				l_neighbour1->setEdgeData(l_edgeIndex+3, l_edgeData);
			}
		}
		else			
		{
			// Record the angle between the first triangle and
			// each other triangle
			std::map<double, std::vector<int> > l_edgeAngles;
			
			for(int l_index = 1; l_index < (*l_iter).second.size(); l_index++)
			{
				double l_angle = PolySurf::getAngleBetweenTriangles(
					(*l_iter).second[0].first, 
					(*l_iter).second[l_index].first,
					l_vertex0,
					l_vertex1);
				
				//XXX If this angle is already in the list then
				// nudge it slightly up.
				if (l_edgeAngles.find(l_angle) != l_edgeAngles.end())
				{
					l_angle += POLYSURF_EPSILON;
				}
				
				l_edgeAngles[l_angle].push_back(l_index);
			}
			
			std::map<double, std::vector<int> >::iterator l_anglesIter = l_edgeAngles.begin();
			
			// Get the first neighbour
			int l_neighbourIndex0 = 0;
			int l_neighbourIndex1;
			
			bool l_isFirstNeighbourOrdered = 
				PolySurf::isTriangleOrdered((*l_iter).second[0].first, l_vertex0, l_vertex1);
			
			bool l_isFirstLoop = true;
			while(l_anglesIter != l_edgeAngles.end())
			{                 
				// Get the next neighbour
				if (!l_isFirstLoop)
				{
					l_anglesIter++;
				}
				else
				{
					l_isFirstLoop = false;
				}
				
				if (l_anglesIter == l_edgeAngles.end())
				{
					l_neighbourIndex1 = 0;
				}
				else
				{
					l_neighbourIndex1 = (*l_anglesIter).second[0];
				}
				
				PolySurf::Triangle* l_neighbour0 = (*l_iter).second[l_neighbourIndex0].first;
				PolySurf::Triangle* l_neighbour1 = (*l_iter).second[l_neighbourIndex1].first;
				
				PolySurf::EdgeData* l_edgeData = new PolySurf::EdgeData();
				m_edgeDataList.push_back(l_edgeData);
				
				l_edgeData->init(l_vertex0, 
					l_vertex1, 
					l_neighbour0, 
					l_neighbour1);
				
				int l_edgeIndex = (*l_iter).second[l_neighbourIndex0].second;
				
				if (PolySurf::isTriangleOrdered(l_neighbour0, l_vertex0, l_vertex1) !=
					l_isFirstNeighbourOrdered)
				{
					l_edgeIndex += 3;
				}
				l_neighbour0->setEdgeData(l_edgeIndex, l_edgeData);
				
				l_edgeIndex = (*l_iter).second[l_neighbourIndex1].second + 3;
				
				if (PolySurf::isTriangleOrdered(l_neighbour1, l_vertex0, l_vertex1) !=
					l_isFirstNeighbourOrdered)
				{
					l_edgeIndex -= 3;
				}
				l_neighbour1->setEdgeData(l_edgeIndex, l_edgeData);
				
				l_neighbourIndex0 = l_neighbourIndex1;               
			}
		}
		l_iter++;
    }  
	
    //
    // Calculate per-vertex normals
    //
    int l_index;
    for(l_index = 0; l_index < m_triangles.size(); l_index++)
    {
        m_triangles[l_index]->getVertex(0)->m_normal += m_triangles[l_index]->getNormal();
        m_triangles[l_index]->getVertex(1)->m_normal += m_triangles[l_index]->getNormal();
        m_triangles[l_index]->getVertex(2)->m_normal += m_triangles[l_index]->getNormal();
    }
    
    for(l_index = 0; l_index < m_vertices.size(); l_index++)
    {
        m_vertices[l_index]->m_normal = m_vertices[l_index]->m_normal.normalize();
    }
}

// Returns the force to be applied.
etVector etLiquidTrisurface::computePointForce(etVector l_position)
{         
	if(!cull(l_position)) {
		
		//	Sleep(400);
		//etVector l_position = toLocal(a_point).getPosition();
		etLiquidTrisurface* l_surface = this;//hapticTrisurface();
		//	l_position = l_position.scale(1/l_surface->getScaleRaw());
		
		
		if (m_isFirstTime)
		{
			m_prevPosition = l_position;
			m_isFirstTime = false;
		}
		
		double l_edgeDistance[] = {0,0,0};          
		bool l_wasFrontFaceTouched = m_isFrontFaceTouched;
		
		//
		// Check for tool moving into a neighboring triangle
		//
		
		int l_cornerIndex;
		int l_neighboursCrossed = 0;    
		
		bool l_justExitedTriangle = false;
		etVector l_triangleExitPoint = m_prevPosition;
		
		if (m_currentTriangle != 0)
		{
			
			// If the force is towards the vertex then check all the triangles
			// and edges attached to the current vertex to see if any contains the 
			// current point, in which case swap to that triangle or edge
			if (m_isForceToVertActive)
			{   
				PolySurf::Vertex* l_sharedVertex = m_currentTriangle->getVertex(m_currentVertex);
				PolySurf::Vertex* l_otherVertex = m_currentTriangle->getVertex((m_currentVertex + 1) % 3);
				// Check neighbouring edges 
				
				
				bool l_checkingOtherSide = false;
				
				std::pair<etVector, etVector> l_line(l_sharedVertex->m_position, l_otherVertex->m_position);
				double l_distanceToEdgeStart = 
					(l_sharedVertex->m_position - closestPointToLine(l_line, l_position)).length();
				
				while(true)
				{
					// Get one of the edges from the current triangle which
					// contains the current vertex, and 
					// get the neighbour for that edge
					//
					// If this is the second pass then get the other edge and neighbour
					PolySurf::Triangle* m_prevTriangle = m_currentTriangle;
					PolySurf::Vertex* l_startEdgeEndVertex;
					const PolySurf::EdgeData* l_edgeData;
					if (!l_checkingOtherSide)
					{
						l_startEdgeEndVertex = m_currentTriangle->getVertex((m_currentVertex+1)%3);
						l_edgeData = m_currentTriangle->getEdgeData(m_currentVertex, l_wasFrontFaceTouched);               
					}
					else
					{
						l_startEdgeEndVertex = m_currentTriangle->getVertex(m_currentVertex);
						l_edgeData = m_currentTriangle->getEdgeData((m_currentVertex+2) % 3, l_wasFrontFaceTouched);
					}
					PolySurf::Triangle* l_neighbour = l_edgeData->getOtherNeighbour(m_currentTriangle);
					
					// Start from the current neighbour and
					// Loop through all triangles attached to this vertex, 
					// checking for an edge or triangle which contains the current position.                
					while(l_neighbour != 0)
					{
						int l_edgeIndex = 0;
						int l_sharedVertIndex = 0;
						// Get the shared vertex's index in neighbour
						while(l_sharedVertIndex < 3 && l_neighbour->getVertex(l_sharedVertIndex) != l_sharedVertex)
						{
							l_sharedVertIndex++;
						}
						
						l_edgeIndex = l_sharedVertIndex;
						
						assert(l_edgeIndex < 3);
						
						// Get the next edge
						const PolySurf::EdgeData* l_nextEdge = l_neighbour->getEdgeData(l_edgeIndex, l_wasFrontFaceTouched);
						if (l_nextEdge->getNeighbour0() == m_prevTriangle ||
							l_nextEdge->getNeighbour1() == m_prevTriangle)
						{             
							l_edgeIndex = (l_edgeIndex+2) % 3;
							l_nextEdge = l_neighbour->getEdgeData(l_edgeIndex, l_wasFrontFaceTouched);
						}               
						
						if (!l_nextEdge->isContactOutsideEdge(l_neighbour, l_wasFrontFaceTouched) && 
							l_nextEdge->getNumNeighbours() == 2)
						{                                                                                 
							std::pair<etVector, etVector> l_line (
								l_neighbour->getVertexPosition(l_edgeIndex),
								l_neighbour->getVertexPosition((l_edgeIndex + 1)%3));                                                                                
							
							//if inside edge, swap to this neighbour                                        
							//XXX add wasFrontFaceTouched toggle.                                        
							if (l_nextEdge->getPerpEdgePlane0().distance(l_position) > 0 &&
								l_nextEdge->getPerpEdgePlane1().distance(l_position) > 0)                                            
							{
								// Check if the tool is past the shared point in direction of this line
								double l_lineLength = (l_line.first - l_line.second).length();
								etVector l_closePoint = closestPointToLine(l_line, l_position);
								double l_distanceToLineEnd = (l_closePoint - l_line.second).length();
								double l_distanceToLineStart = (l_closePoint - l_line.first).length();
								
								if (l_edgeIndex != l_sharedVertIndex)
								{
									double l_temp = l_distanceToLineEnd;
									
									l_distanceToLineEnd = l_distanceToLineStart;
									l_distanceToLineStart = l_temp;
								}                           
								
								// If in vert and inside this edge
								//      if on the correct side of one of this edges nighbours then
								//          swap into that neighbour
								if (l_distanceToLineEnd < l_lineLength ||
									l_distanceToLineEnd < l_distanceToLineStart)                                
								{
									// Only swap to a neighbour if the tool is on the correct
									// side of that neighbour                                                
									if ((triangleToPlane(l_neighbour).distance(l_position) < 0) == 
										l_wasFrontFaceTouched)
									{                                                                     
										m_currentTriangle = l_neighbour;                                    
										
										// The tool is inside this edge, so
										// the force won't be towards the vertex
										m_isForceToVertActive = false;                                    
										
										break;
									}
									else
									{
										PolySurf::Triangle * l_otherNeighbour = l_nextEdge->getOtherNeighbour(l_neighbour);
										if ((triangleToPlane(l_otherNeighbour).distance(l_position) < 0) == 
											l_wasFrontFaceTouched)
										{                                        
											m_currentTriangle = l_otherNeighbour;
											
											m_isForceToVertActive = false;
											
											break;
										}
									}                                               
								}                            
							}                                           
						} // is contact outside edge
						
						// If didn't swap to the edge, check for between
						// perp edge planes of the current neighbour for the edges
						// which attach to the current vertex
						const PolySurf::EdgeData* l_edge = l_neighbour->getEdgeData(l_sharedVertIndex, l_wasFrontFaceTouched);
						double l_distance; 
						
						if (l_edge->getNeighbour0() == l_neighbour)
						{
							l_distance = l_edge->getPerpEdgePlane0().distance(l_position);
						}
						else                    
						{
							l_distance = l_edge->getPerpEdgePlane1().distance(l_position);
						}
						if (l_distance <= 0)
						{                        
							l_edge = l_neighbour->getEdgeData((l_sharedVertIndex+2)%3, l_wasFrontFaceTouched);
							if (l_edge->getNeighbour0() == l_neighbour)
							{
								l_distance = l_edge->getPerpEdgePlane0().distance(l_position);
							}
							else                    
							{
								l_distance = l_edge->getPerpEdgePlane1().distance(l_position);
							}
							
							if (l_distance <= 0)
							{                    
								// Inside the two edges so swap to this neighbour
								m_currentTriangle = l_neighbour;                                                         
								
								m_isForceToVertActive = false;
								break;                                    
							}
						}
						
						// Exit the loop if all triangles havre been
						// checked
						if (l_neighbour == m_currentTriangle)
						{
							break;
						}
						
						// Get the next neighbouring triangle
						m_prevTriangle = l_neighbour; 
						double l_distanceCheck = 0;
						if (l_neighbour == l_nextEdge->getNeighbour0())
						{
							l_neighbour = l_nextEdge->getNeighbour1(); 
							if (l_neighbour == 0)
							{
								l_distanceCheck = l_nextEdge->getEdgePlane().distance(l_position);
							}
						}
						else
						{
							l_neighbour = l_nextEdge->getNeighbour0();
							if (l_neighbour == 0)
							{
								l_distanceCheck = -l_nextEdge->getEdgePlane().distance(l_position);
							}
						}
						
						// If this is the last edge and the tool is past it then
						// exit the surface
						if (l_neighbour == 0 && l_distanceCheck > 0)                    
						{
							m_currentTriangle = 0;
							m_isForceToVertActive = false;
						}
					} //while 
					
					// If neighbour check succeded or went through
					// all neighbours don't check from other side.
					if (l_neighbour != 0 || !m_isForceToVertActive || l_checkingOtherSide)
					{
						break;
					}
					l_checkingOtherSide = true;
				}
			}
			// Check for crossing over an edge into a neighbouring triangle,
			// or for entering a vertex.
			else if (!m_isForceToVertActive)
			{            
				for(l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
				{   
					const PolySurf::EdgeData* l_edgeData = m_currentTriangle->getEdgeData(l_cornerIndex, l_wasFrontFaceTouched);
					
					// If on inside of an edge than use a perp edge plane                
					if (!l_edgeData->isContactOutsideEdge(m_currentTriangle, l_wasFrontFaceTouched))                    
					{                                                        
						etVector l_vector0 = m_currentTriangle->getVertexPosition(l_cornerIndex);
						etVector l_vector1 = m_currentTriangle->getVertexPosition((l_cornerIndex+1)%3);
						
						PolySurf::Plane l_perpEdgePlane(
							l_vector0,
							l_vector1,
							l_vector0 + m_currentTriangle->getNormal());
						
						// If the tool is inside the edge plane then don't
						// exit this edge                    
						PolySurf::Triangle* l_neighbour = l_edgeData->getNeighbour0();
						double l_distance1 = l_edgeData->getPerpEdgePlane0().distance(l_position);
						double l_distance0 = l_edgeData->getPerpEdgePlane1().distance(l_position);
						if (m_currentTriangle == l_neighbour)
						{
							l_neighbour = l_edgeData->getNeighbour1();
							double l_temp = l_distance0;
							l_distance0 = l_distance1;
							l_distance1 = l_temp;
						}                 
						if (l_distance0 < 0)
						{                       
							continue;
						}
						else // check for entering neighbour
						{                      
							if (l_neighbour == 0)
							{                         
								m_currentTriangle = 0;
								break;
							}
							
							// If outside neighbour, check for inside edge
							if (l_distance1 > 0)
							{                            
								// get force towards edge, or point
								std::pair<etVector, etVector> l_line(l_vector0, l_vector1);
								etVector l_projToLine = closestPointToLine(l_line, l_position);
								
								double l_edgeLength = (l_vector0 - l_vector1).length();
								int l_currentVertex = -1;                            
								
								if ((l_vector0 - l_projToLine).length() > l_edgeLength )
								{
									// Crossed over line end
									l_currentVertex = (l_cornerIndex+1)%3;                                
								}
								else if (((l_vector1 - l_projToLine).length() > l_edgeLength))
								{
									// Crossed over line start
									l_currentVertex = l_cornerIndex;                              
								}                          
								
								// Test for moving past edge end, into the vertex.                                                     
								if (l_currentVertex != -1)
								{                              
									m_isForceToVertActive = true;
									m_currentVertex = l_currentVertex;                                
									
									break;                               
								}
							}                       
						}
					}
				}
			}
			
			// Check for crossing an edge plane into a neighbouring triangle
			
			PolySurf::Triangle* l_neighbour = m_currentTriangle;
			int l_neighboursCrossed = 0;
			bool l_isInNeighbour;
			
			if (!m_isForceToVertActive)
			{
				while(l_neighbour != 0)
				{
					l_isInNeighbour = false;  
					l_neighboursCrossed++;                            
					for(l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
					{                       
						const PolySurf::EdgeData* l_edgeData = l_neighbour->getEdgeData(l_cornerIndex, l_wasFrontFaceTouched);
						
						assert(l_edgeData != 0);              
						
						l_edgeDistance[l_cornerIndex] = l_edgeData->getEdgePlane().distance(l_position);
						
						// If the angle between this and the neighbour is acute
						// and the tool is on the outside of the edge, then
						// do not cross to neighbour through this edge
						if (l_edgeData->isAcute() && 
							l_edgeData->isContactOutsideEdge(l_neighbour, l_wasFrontFaceTouched) && 
							l_neighboursCrossed == 1)
						{   
							// Get an edge plane perpendicular to the triangle
							PolySurf::Plane l_perpEdgePlane(
								l_neighbour->getVertexPosition(l_cornerIndex),
								l_neighbour->getVertexPosition((l_cornerIndex+1)%3),
								l_neighbour->getVertexPosition(l_cornerIndex) + 
								l_neighbour->getNormal());
							
							// If the tool is past the edge plane then exit the surface
							if (l_perpEdgePlane.distance(l_position) > 0)
							{
								// Catch re-intersections                        
								double l_u;
								l_justExitedTriangle = true;                
								if (!triangleToPlane(l_neighbour).getIntersection(l_position, l_position + l_neighbour->getNormal(), l_u, l_triangleExitPoint))
								{
									l_triangleExitPoint = m_prevPosition;                            
								}
								
								l_neighbour = 0;   
								l_isInNeighbour = true;
								
								break;
							}
							
							// Don't check for crossing edge plane
							continue;
						}
						
						// Check for crossed out of the triangle
						// Note that the edge plane normal points towards neighbour 1
						// so points inside neighbour 0 will have negative distance               
						if (l_neighbour == l_edgeData->getNeighbour0() && l_edgeDistance[l_cornerIndex] > 0)
						{
							l_neighbour = l_edgeData->getNeighbour1();
							l_isInNeighbour = true;
						}
						else if (l_neighbour == l_edgeData->getNeighbour1() && l_edgeDistance[l_cornerIndex] < 0)
						{
							l_neighbour = l_edgeData->getNeighbour0();
							l_isInNeighbour = true;
						} 
						
						if (l_isInNeighbour)  
						{
							// If the normal flips when crossing this edge
							// then reverse which face is touched
							if (l_edgeData->areNeighbourNormalsFlipped())
							{
								l_wasFrontFaceTouched = !l_wasFrontFaceTouched;
							}
							// Don't check remaining edges
							break;
						}  
					}// for(.., cornerIndex < 3, ..)           
					
					// If the tool in inside the edge planes
					// of a wrong side triangle then allow the 
					// side swap test to exit the surface
					if (!l_isInNeighbour && 
						l_neighbour != 0 &&
						l_neighboursCrossed > 1 &&
						(triangleToPlane(l_neighbour).distance(l_position) < 0) != 
						l_wasFrontFaceTouched)
					{
						l_isInNeighbour = true;
						m_currentTriangle = l_neighbour;
					}
					
					if (!l_isInNeighbour ||
						l_neighbour == 0 ||
						l_neighbour == m_currentTriangle ||
						l_neighboursCrossed > POLYSURF_NEIGHBOUR_CROSS_LIMIT ||
						(triangleToPlane(l_neighbour).distance(l_position) < 0) == 
						l_wasFrontFaceTouched)
					{
						break;
					}   
				} // while     
			} // if
			
			if (l_isInNeighbour)
			{
				m_currentTriangle = l_neighbour;
			}
			
		}// m_currentTriangle != 0
		
		
		//
		// Exit triangle if the cursor has swapped sides
		//
		PolySurf::Plane l_trianglePlane;
		double l_distanceToTri;
		
		if (m_currentTriangle != 0)
		{       
			
			// If l_position is in behind the current triangle
			//      then m_isFrontFaceTouched = true 
			//      else m_isFrontFaceTouched = false
			l_trianglePlane = PolySurf::Plane(m_currentTriangle->getNormal(), m_currentTriangle->getVertexPosition(0));
			l_distanceToTri = l_trianglePlane.distance(l_position);
			
			if (!m_isForceToVertActive)
			{
				m_isFrontFaceTouched = (l_distanceToTri < 0);
				
				// If l_position has swapped sides of the surface then exit the surface             
				if (m_isFrontFaceTouched != l_wasFrontFaceTouched && !m_isForceToVertActive)
				{        
					double l_u;
					l_justExitedTriangle = true;                
					if (!l_trianglePlane.getIntersection(l_position, m_prevPosition, l_u, l_triangleExitPoint))
					{
						l_triangleExitPoint = m_prevPosition;                   
					}
					
					m_currentTriangle = 0;           
				}        
			}
		}
		
		//
		// Check for intersection with a new triangle
		//
		
		// If there is no current triangle then check for intersection
		// with a new triangle by colliding with the OctTree
		bool l_isNewTriangle = false;
		if (m_currentTriangle == 0)
		{
			l_isNewTriangle = true;
			
			if (l_surface->getOctTree())
			{   
				// If justExitedTriangle then 
				//      only intersect with the side of the surface which was touched before
				// else intersect with all triangles
				if (l_justExitedTriangle)
				{
					ReintersectingTriangleTest l_triangleTest(l_wasFrontFaceTouched, l_position);
					
					m_currentTriangle = l_surface->getOctTree()->getCollisionIf(l_triangleExitPoint, l_position, l_triangleTest);
				}
				else
				{	
					m_currentTriangle = l_surface->getOctTree()->getCollision(m_prevPosition, l_position);
				}            
			}
			
			// Check if the touched face is active.                
			if (m_currentTriangle)
			{
				l_trianglePlane = PolySurf::Plane(m_currentTriangle->getNormal(),
					m_currentTriangle->getVertexPosition(0));
				l_distanceToTri = l_trianglePlane.distance(l_position);      
				m_isFrontFaceTouched = (l_distanceToTri < 0);
				
				if ( (m_isFrontFaceTouched && !l_surface->isFrontFaceTouchEnabled()) ||
					(!m_isFrontFaceTouched && !l_surface->isBackFaceTouchEnabled()) )
				{                         
					m_currentTriangle = 0;                  
				}
			}
		}
		
		// Update the prev position now that we are done with it
		m_prevPosition = l_position;
		
		//
		// If still no triangle then  return no force       
		//
		if (m_currentTriangle == 0)
		{ 
			l_surface->setCurrentTriangleRaw(0);
			l_surface->setToolContactPositionRaw(etVector());
			doFriction = false;

			return etVector();
		}
		
		//
		// Generate the force for the current triangle
		//
		
		// Force direction set to triangle normal
		etVector l_forceDirection = l_trianglePlane.normal();
		
		bool l_isForceTowardsEdge = false;       
		
		// If m_isForceToVertActive then set force direction towards 
		// the current vertex
		if (m_isForceToVertActive)
		{
			l_forceDirection = m_currentTriangle->getVertexPosition(m_currentVertex) - l_position;
			l_distanceToTri = -l_forceDirection.length();
			l_forceDirection = l_forceDirection.scale(-1/l_distanceToTri);
			l_isForceTowardsEdge = true;
		}
		else
		{
			// If we are on the inside of a very acute angle and 
			//    the projection of l_psoiton onto the triangles plane is outside the triangle then
			//          set the force to be towards the edge 
			
			//recalculate the edge distances for the new triangle
			
			for(l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
			{   
				const PolySurf::EdgeData* l_edgeData = m_currentTriangle->getEdgeData(l_cornerIndex, m_isFrontFaceTouched);
				
				if (l_edgeData != 0)
				{
					l_edgeDistance[l_cornerIndex] = l_edgeData->getEdgePlane().distance(l_position);
				}
			}       
			
			for(int l_edgeIndex = 0; l_edgeIndex < 3; l_edgeIndex++)
			{ 
				const PolySurf::EdgeData* l_edgeData = m_currentTriangle->getEdgeData(l_edgeIndex, m_isFrontFaceTouched);
				if (l_edgeData == 0)
				{
					continue;
				}
				
				// If the angle between the edges neighbours is very acute and 
				// the tool is touching the concave side of the join then
				//      For the current triangle and it's neighbouring triangle on this edge
				//          if the position's projection onto each triangles plane 
				//          is outside of the triangle then 
				//              set the force to point to the edge
				
				if (!l_edgeData->isContactOutsideEdge(m_currentTriangle, m_isFrontFaceTouched))
				{
					
					// Get the edge  points
					etVector l_edgePoint0 = m_currentTriangle->getVertexPosition(l_edgeIndex);
					etVector l_edgePoint1 = m_currentTriangle->getVertexPosition((l_edgeIndex+1)%3);
					
					// Get a plane along the edge perpendicular to the current triangle
					PolySurf::Plane l_edgePerpPlane(
						l_edgePoint0,
						l_edgePoint1,
						l_edgePoint0 + m_currentTriangle->getNormal());
					
					// Check for crossing out of the current triangle
					if (l_edgePerpPlane.distance(l_position) > 0)
					{
						PolySurf::Triangle* l_neighbour = l_edgeData->getNeighbour0();
						if (l_neighbour == m_currentTriangle)
						{
							l_neighbour = l_edgeData->getNeighbour1();                   
						}
						
						if (l_neighbour == 0)
						{
							continue;
						}
						
						// If neighbours have normals flipped
						// then reverse the direction of the edge
						if (l_edgeData->areNeighbourNormalsFlipped())
						{
							etVector l_temp = l_edgePoint1;
							l_edgePoint1 = l_edgePoint0;
							l_edgePoint0 = l_temp;
						}
						
						// Get a plane along the edge perp to the neighbouring triangle
						l_edgePerpPlane = PolySurf::Plane(
							l_edgePoint1,
							l_edgePoint0,
							l_edgePoint0 + l_neighbour->getNormal());
						
						// If the position is outside the neighbouring triangle then
						//      if the point is within the edge
						//      the force to be towards the edge
						double l_distanceToEdgePlane = l_edgePerpPlane.distance(l_position);
						if (l_distanceToEdgePlane > 0)
						{
							// Get a line for the edge
							std::pair<etVector, etVector> l_line(
								m_currentTriangle->getVertexPosition(l_edgeIndex),
								m_currentTriangle->getVertexPosition((l_edgeIndex+1)%3));
							
							double l_lineLength = (l_line.first - l_line.second).length();
							etVector l_closePoint = closestPointToLine(l_line, l_position);
							double l_distanceToLineEnd = (l_closePoint - l_line.second).length();
							double l_distanceToLineStart = (l_closePoint - l_line.first).length();
							
							// if the point on the line is between the
							//    the start and end of the line then
							//      set the force to be towards the edge
							if (l_distanceToLineEnd < l_lineLength && 
								l_distanceToLineStart < l_lineLength)
							{
								// Set force to be towards the line
								etVector l_pointOnLine = closestPointToLine(l_line, l_position);
								l_forceDirection = l_pointOnLine - l_position;                    
								
								l_isForceTowardsEdge = true;           					
								
								l_distanceToTri = -l_forceDirection.length();
								l_forceDirection = l_forceDirection.scale(-1/l_distanceToTri);
								
								break;
							}
						}
					}
				}
			}
		}
		
		// Interpolate force if the force has not been set as towards an edge
		if (!l_isForceTowardsEdge)
		{   
			//
			// Interpolate force to the neighbouring triangles
			//
			
			// Set the interpolation range for each edge to be 
			//      the distance from the center of the edge to the opposite corner * POLYSURF_INTERP_DISTANCE
			//      
			// if the tool is on the outside of an acute 
			//      angle then don't interp that edge.
			double l_interpRange[3];
			bool l_isEdgeInterpolated[] = {true, true, true};       
			
			for(l_cornerIndex = 0;  l_cornerIndex < 3; l_cornerIndex++)
			{     
				
				const PolySurf::EdgeData* l_edgeData = m_currentTriangle->getEdgeData(l_cornerIndex, m_isFrontFaceTouched);
				
				// Don't interp if on the outside of an acute angled edge,
				// or on the inside of any angle
				if (l_edgeData == 0 ||
					l_edgeData->isAcute() ||                
					(!l_edgeData->isContactOutsideEdge(m_currentTriangle, m_isFrontFaceTouched)))
				{
					l_isEdgeInterpolated[l_cornerIndex] = false;               
				}
				
				int l_nextCorner = (l_cornerIndex + 1) % 3;
				int l_oppositeCorner = (l_cornerIndex + 2) % 3;
				
				// XXX could pre-calculate edge lengths
				std::pair<etVector, etVector> l_line(
					m_currentTriangle->getVertexPosition(l_cornerIndex),
					m_currentTriangle->getVertexPosition(l_nextCorner));
				
				etVector l_oppositeCornerPos = m_currentTriangle->getVertexPosition(l_oppositeCorner);
				l_interpRange[l_cornerIndex] = 
					(closestPointToLine(l_line, l_oppositeCornerPos) - l_oppositeCornerPos).length()
					* POLYSURF_INTERP_DISTANCE;                     
			}
			
			bool l_isForceInterpolated = false;
			
			double l_interpFactor[3] = {0,0,0};
			
			// Calculate the interpolation factor for each edge
			for(l_cornerIndex = 0;  l_cornerIndex < 3; l_cornerIndex++)
			{                      
				if (l_isEdgeInterpolated[l_cornerIndex] &&
					fabs(l_edgeDistance[l_cornerIndex]) < l_interpRange[l_cornerIndex] &&
					m_currentTriangle->getNeighbour(l_cornerIndex, m_isFrontFaceTouched) != 0)
				{
					l_interpFactor[l_cornerIndex] = 
						(l_interpRange[l_cornerIndex] - 
						fabs(l_edgeDistance[l_cornerIndex])) /
						l_interpRange[l_cornerIndex];
				}           
			}
			
			// Apply the interp factor for each edge to the force direction
			for(l_cornerIndex = 0;  l_cornerIndex < 3; l_cornerIndex++)
			{
				if (l_interpFactor[l_cornerIndex] > 0)
				{
					const PolySurf::EdgeData* l_edgeData = m_currentTriangle->getEdgeData(l_cornerIndex, m_isFrontFaceTouched);
					PolySurf::Triangle* l_neighbour = m_currentTriangle->getNeighbour(l_cornerIndex, m_isFrontFaceTouched);
					etVector l_neighbourForce;
					
					l_neighbourForce = m_currentTriangle->getNeighbour(l_cornerIndex, m_isFrontFaceTouched)->getNormal();
					
					if (l_edgeData->areNeighbourNormalsFlipped())
					{
						l_neighbourForce = l_neighbourForce.scale(-1);
					}                                                                  
					l_forceDirection += l_neighbourForce.scale(l_interpFactor[l_cornerIndex]);
					
					l_isForceInterpolated = true;
				}            
			}
			
			if (l_isForceInterpolated)
			{
				l_forceDirection = l_forceDirection.normalize();
			}       
		}       
		
		//etVector l_force = l_forceDirection.scale(-l_distanceToTri);
		l_surface->setCurrentTriangleRaw(m_currentTriangle);
		
		etVector l_force;
		
		//etLiquidTrisurface* l_surface = this;	
		
		if (m_isFirstTime) {
			l_prevPosition = l_position;
			m_prevPosition = l_position;
			m_isFirstTime = false;
			return etVector(0,0,0);
		}
		
		etVector distance = l_position - l_prevPosition;
		m_isFirstTime = false;
		
		// Update the prev position now that we are done with it
		l_prevPosition = l_prevPosition + distance.scale(0.1);
		m_prevPosition = l_position;
		
		if (distance.length() < 0.01)
			return etVector(0,0,0);
		else
			l_force = -distance.scale(viscosity);

		return l_force;
	}
	else {
		// Object is culled away.
		m_isFirstTime = true;
		return etVector(0,0,0);
	}
}

void etLiquidTrisurface::init(bool a_isVertexMergeEnabled)
{   
	// Disable forces
	bool l_isFrontFaceEnabled = isFrontFaceTouchEnabled();
	bool l_isBackFaceEnabled = isBackFaceTouchEnabled();
	
	disableFrontFaceTouch();
	disableBackFaceTouch();
	
	initOctTree();
	
	// Merge vertices
	if (a_isVertexMergeEnabled)
	{
		MergeVerticesInNode l_merge;
		if (m_octTree)
		{
			m_octTree->forEachItem(l_merge);        
		}       
	}
	
	processEdges();
	
	if (l_isFrontFaceEnabled)
	{
		enableFrontFaceTouch();
	}
	if (l_isBackFaceEnabled)
	{
		enableBackFaceTouch();
	}
}


bool etLiquidTrisurface::cull(etVector l_position)
{         
	bool l_isCulled = false;
	etLiquidTrisurface* l_surface = this;
	
	if (m_isFirstTime) {
		m_prevPosition = l_position;
		l_prevPosition = l_position;
	}

	//Check if the latest PHANTOM movement is inside the octree of the object.
	std::pair<etVector, etVector> l_line(m_prevPosition,l_position);
	
	if (!isLineInBox(l_line, l_surface->getOctTree()->getCenter(),l_surface->getOctTree()->getSize())) {
		l_isCulled = true;
	}
	
	if (l_isCulled)
	{
		return true;
	}
	return false;
}

void etLiquidTrisurface::setViscosity(double v) {
	if(v<0)
		v = 0;
	else if(v>1)
		v = 1;
	viscosity = v * 500;
}