#ifndef LIQUID_TRISURFACE
#define LIQUID_TRISURFACE

// Includes:
//	#include <et/vector.h>
//#include <et/variable.h>
//#include <et/orientation.h>

#include "object.h"
#include <vector>
#include <map>



// Forward declarations
class etOctTree;

namespace PolySurf{
    struct Vertex;
    class Plane;
    class Triangle;
    class EdgeData;
}

//
// Class:
//     etHapticTrisurface
//
// Description:
//     A triangulated surface class
//
//     Triangles and vertices can be loaded using the create method.
//     Alternately,
//       use setVertices to load the vertices,
//       addTriangle for each triangle
//       and then call init().
//
// Exceptions:
//     None
//
// Notes:
//     None
//
// See Also Classes:
//      OctTree,
//      Plane,
//      Vertex,
//      Triangle,
//      EdgeData
//
// See Also Functions:
//     None
//
class etLiquidTrisurface : public etObject
{
    public:
    
      typedef PolySurf::Triangle* triangle_ref;

      // Surface point of the tool in parent coordinates.
      ET_SHARED_VARIABLE(etVector,ToolContactPosition);

      // Scaling factor fro the surface.
      ET_SHARED_VARIABLE(double,Scale);
      
      // Stiffness of the surface.
      ET_SHARED_VARIABLE(double,Stiffness);
    
      // Reference to the triangle the tool is touching.
      ET_SHARED_VARIABLE(triangle_ref, CurrentTriangle);  

      // Highlight the current triangle.
      ET_FLAG(DrawCurrentTriangle);

      // Allow the front of the surface to be touched.
      ET_FLAG(FrontFaceTouch);

      // Allow the back of the surface to be touched.
      ET_FLAG(BackFaceTouch);

    public:      
                         
        // Constructor.
        etLiquidTrisurface();
   
        // Destructor.
        virtual ~etLiquidTrisurface();


        //
        // Initialisation
        //

        // Create initialises a surface.
        // 
        // a_coordinates should list the triangles as 
        // groups of indices into the vertices list,
        // each triangle should be three inidices followed
        // by a -1, similar to a VRML indexedFaceSet.
        void create(std::vector<etVector> a_vertices, 
                    std::vector<int> a_coordinates, 
                    bool a_isVertexMergeEnabled = false);

        // Initialises the OctTree and the edge data structures.        
        void init(bool a_isVertexMergeEnabled = false);

        // Replaces the vertex list with a_vertices.
        void setVertices(const std::vector<etVector>& a_vertices);

        // Adds a triangle using indices into the vertex list.
        // The force structures won't be updated until init is called.
        void addTriangle(int a_index0, int a_index1, int a_index2);

        // Adds a triangle, using Vertex pointers.
        // assumes the Vertices are already handled.
        void addTriangle(PolySurf::Vertex* a_point0, 
                         PolySurf::Vertex* a_point1, 
                         PolySurf::Vertex* a_point2);
       
        //Retruns the list of triangles in the surface.
        inline const std::vector<PolySurf::Triangle*>& triangles() const
        {
            return m_triangles;
        }
         
        // Returns the oct-tree used for culling the surface.
        const etOctTree* getOctTree() const
        {
            return m_octTree;
        }

		etVector computePointForce(etVector l_position);
				
		bool cull(etVector l_position);

		void setViscosity(double v);

    protected:

        // Remove all edgeDatas.
        void clearEdgeData();

        // Remove all triangles and vertices
        void clearTrianglesAndVertices();

         // Calculate edge planes for all edges, called by init.
        void processEdges();
                       
        // Initialise the oct tree, called by init.
        void initOctTree();      

    private:
    
        // Copy constructor.
        etLiquidTrisurface(const etLiquidTrisurface& other);

        // Assignment operator.
        etLiquidTrisurface& operator=(const etLiquidTrisurface& other);

        // Returns a new copy of the surface.
        etLiquidTrisurface* clone() const;
          
        // The triangles list.
        std::vector<PolySurf::Triangle* > m_triangles;
    
        // The vertex list.
        std::vector< PolySurf::Vertex* > m_vertices;

        // Octtree used to cull triangles in collision detection.
        etOctTree* m_octTree;

        // List of edge datas,
        std::vector<PolySurf::EdgeData*> m_edgeDataList;
		
		etVector m_prevPosition;
		etVector l_prevPosition;
		bool m_isFrontFaceTouched;
		bool m_isForceToVertActive;
		int m_currentVertex;
		PolySurf::Triangle* m_currentTriangle;
		bool m_isFirstTime;
		double viscosity;
		
};

#endif
