#include "cssysdef.h"
#include "iengine/mesh.h"

#include "octree.h"

/** Constructor. Creates the root of the octree. (Recursively creates the whole tree.) */
WorldOctree::WorldOctree(std::vector<csBox3> boundBoxes,std::vector<double> minMaxVec) {
	printf("Creating world octree...\n");
	root = NULL;
	boxes = boundBoxes;
	nbrBoxes = boundBoxes.size();
	minMax = minMaxVec;
	std::list<int> boxIndex;

	for(int i=0;i<nbrBoxes;i++)
		boxIndex.push_back(i);
	root = new Node(this,minMaxVec[0],minMaxVec[1],minMaxVec[2],minMaxVec[3],-minMaxVec[5],-minMaxVec[4],boxIndex,0);
}

/** Destructor. */
WorldOctree::~WorldOctree() {
	if (root) delete root;
}

/** Returns the bounding box with index nbr. */
csBox3 WorldOctree::getBox(int nbr) const{
	return boxes[nbr];
}

/** Returns a index list of objects close to the PHANTOM position. */
std::list<int> WorldOctree::getNearbyObjects(double x, double y, double z) {
	std::list<int> objectList = root->getNearbyObjects(x,y,z);
	return objectList;
}

////////////////////////////////////////////////// Class Node //////////////////////////////////////////////////////////////////


/** Constructor. Recursively creates an octree. If the node contains no objects or if it is on the max depth no children are added. */
Node::Node(WorldOctree* o, double minX,double maxX,double minY,double maxY,double minZ,double maxZ,std::list<int> index,int d) {
	octree = o;
	depth = d;
	this->minX = minX;
	this->maxX = maxX;
	this->minY = minY;
	this->maxY = maxY;
	this->minZ = minZ;
	this->maxZ = maxZ;
	this->avgX = (minX+maxX)/2;
	this->avgY = (minY+maxY)/2;
	this->avgZ = (minZ+maxZ)/2;
	csBox3 checkBox = csBox3(minX,minY,minZ,maxX,maxY,maxZ);
	std::list<int>::iterator box;
	int count = 0;

	for (box = index.begin(); box != index.end(); ++box) {
		count++;
		csBox3 theBox = octree->getBox(*box);

		if(checkBox.Overlap(theBox) || checkBox.Adjacent2(theBox) !=-1)
			objectIndex.push_back(*box);
	}
	count = 0;

	if (!objectIndex.empty() && depth!=MAX_DEPTH) {
		children.resize(8);
		children[0] = new Node(octree,minX,avgX,minY,avgY,minZ,avgZ,objectIndex,depth+1); 
		children[1] = new Node(octree,minX,avgX,minY,avgY,avgZ,maxZ,objectIndex,depth+1);
		children[2] = new Node(octree,avgX,maxX,minY,avgY,avgZ,maxZ,objectIndex,depth+1);
		children[3] = new Node(octree,avgX,maxX,minY,avgY,minZ,avgZ,objectIndex,depth+1);
		children[4] = new Node(octree,minX,avgX,avgY,maxY,minZ,avgZ,objectIndex,depth+1);
		children[5] = new Node(octree,minX,avgX,avgY,maxY,avgZ,maxZ,objectIndex,depth+1);
		children[6] = new Node(octree,avgX,maxX,avgY,maxY,avgZ,maxZ,objectIndex,depth+1);
		children[7] = new Node(octree,avgX,maxX,avgY,maxY,minZ,avgZ,objectIndex,depth+1);
	}		
}

/** Destructor. */
Node::~Node() {
	if (children.size() !=0) {
		for(int i=0;i<8;i++) {
			delete children[i];
		}
	}
}

/** Returns an index list with objects close to the PHANTOM position. Calls the getNearbyObjects for its children recursively. */
std::list<int> Node::getNearbyObjects(double x, double y, double z) const {
	if (depth == MAX_DEPTH)
		return objectIndex;
	else if (children.size() == 0)
		return objectIndex;
	else {
		if (y < avgY) {
			if (x < avgX) {
				if (z < avgZ)
					return children[0]->getNearbyObjects(x,y,z);
				else
					return children[1]->getNearbyObjects(x,y,z);
			}
			else {
				if (z < avgZ)
					return children[3]->getNearbyObjects(x,y,z);
				else
					return children[2]->getNearbyObjects(x,y,z);
			}
		}
		else {
			if (x < avgX) {
				if (z < avgZ)
					return children[4]->getNearbyObjects(x,y,z);	
				else
					return children[5]->getNearbyObjects(x,y,z);	
			}
			else {
				if (z < avgZ)
					return children[7]->getNearbyObjects(x,y,z);
				else
					return children[6]->getNearbyObjects(x,y,z);
			}
		}
	}	
}

