// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// oct_tree.cpp -- contains the User implementation
//
// $Revision: #2 $
// $Change: 1931 $
//

//XXX yuck!  turn off warning about symbol names being too long
#pragma warning(disable:4786)

// Includes
#include <et/common.h>
#include <GL/gl.h>

#include "octtree.h" //<et/objects/oct_tree.h>
#include <et/objects/plane.h>

#include <utility>

const int OCT_TREE_MAX_DEPTH = 5;
const int OCT_TREE_MAX_TRIANGLES_PER_NODE = 2;

// Expand max bounds
const double OCTTREE_BOUNDS_ALLOWANCE = 1.1;

const int OCTANT_X_MASK = 0x4;
const int OCTANT_Y_MASK = 0x2;
const int OCTANT_Z_MASK = 0x1;

const int OCT_TREE_BOX_EDGES[24] = {
            etOctTree::OIT_LBB, etOctTree::OIT_RBB, // Left to right lines
            etOctTree::OIT_LBF, etOctTree::OIT_RBF,
            etOctTree::OIT_LTB, etOctTree::OIT_RTB,
            etOctTree::OIT_LTF, etOctTree::OIT_RTF,
            etOctTree::OIT_LBB, etOctTree::OIT_LTB, // Bottom to Top lines
            etOctTree::OIT_LBF, etOctTree::OIT_LTF,
            etOctTree::OIT_RBB, etOctTree::OIT_RTB,
            etOctTree::OIT_RBF, etOctTree::OIT_RTF,
            etOctTree::OIT_LBB, etOctTree::OIT_LBF, // Back to Front lines
            etOctTree::OIT_LTB, etOctTree::OIT_LTF,
            etOctTree::OIT_RBB, etOctTree::OIT_RBF,
            etOctTree::OIT_RTB, etOctTree::OIT_RTF
 };

const etVector OCT_TREE_BOX_NORMALS[6] = {
    etVector(0,0,1),
    etVector(0,0,-1),
    etVector(0,1,0),
    etVector(0,-1,0),
    etVector(1,0,0),
    etVector(-1,0,0)
};

const etVector OCT_TREE_BOX_CORNERS[8] = {
    etVector(-1,-1,-1),    //OIT_LBB  
    etVector(-1,-1, 1),    //OIT_LBF 
    etVector(-1, 1,-1),    //OIT_LTB 
    etVector(-1, 1, 1),    //OIT_LTF 
    etVector( 1,-1,-1),    //OIT_RBB 
    etVector( 1,-1, 1),    //OIT_RBF 
    etVector( 1, 1,-1),    //OIT_RTB 
    etVector( 1, 1, 1)     //OIT_RTF 
};


namespace{
    
    // Triangle test which always returns true
    class DefaultTriangleTest : public std::unary_function<etOctTree::item_type, bool>
    {
        operator()(PolySurf::Triangle* a_triangle)
        {
            return true;
        }
    };
}

// Check for intersection between a box and a line
bool isLineInBox(std::pair<etVector, etVector> a_line,
                 const etVector& a_boxCenter,
                 const etVector& a_boxSize)
{
    etVector l_halfBoxSize = a_boxSize.scale(0.5);
    double l_farPlanes[] ={
            a_boxCenter.x() + l_halfBoxSize.x(),
            a_boxCenter.y() + l_halfBoxSize.y(),
            a_boxCenter.z() + l_halfBoxSize.z() };
    double l_nearPlanes[] = { 
            a_boxCenter.x() - l_halfBoxSize.x(),
            a_boxCenter.y() - l_halfBoxSize.y(),
            a_boxCenter.z() - l_halfBoxSize.z() };
    double l_start[3] = {
            a_line.first.x(),
            a_line.first.y(),
            a_line.first.z() };
    double l_diff[3] = {
            a_line.second.x() - a_line.first.x(),
            a_line.second.y() - a_line.first.y(),
            a_line.second.z() - a_line.first.z() };

    double l_tNear = -100000; //XXX should use a constant for large number
    double l_tFar = 100000;

    for(int i = 0; i < 3; i++)
    {
        // If the line doesn't vary in this axis then
        // just test if the start point is between the test planes
        if (l_diff[i] == 0)
        {
            if (l_start[i] > l_farPlanes[i] ||
                l_start[i] < l_nearPlanes[i])
            {
                return false;
            }            
        }
        else
        {
            double t1 = (l_nearPlanes[i] - l_start[i]) / l_diff[i];
            double t2 = (l_farPlanes[i] - l_start[i]) / l_diff[i];
            if (t1 > t2)
            {
                double l_temp = t1;
                t1 = t2;
                t2 = l_temp;
            }
            if (t1 > l_tNear)
            {
                l_tNear = t1;
            }
            if (t2 < l_tFar)
            {
                l_tFar = t2;
            }
            if (l_tNear > l_tFar || l_tFar < 0)
            {
                return false;
            }
        }
    }    
    return true;
}

// Returns true if the triangle overlaps the box
bool isTriangleInBox(etOctTree::item_type a_triangle, const etVector& a_boxCenter, const etVector& a_boxSize)
{        
    etVector l_intersection;
    double l_intersectionDistance;
    etVector l_halfBoxSize = a_boxSize.scale(0.5);

    // Check for intersection between the triangles edges
    // and the sides of the box
    for(int l_sideIndex = 0; l_sideIndex < 6; l_sideIndex++)
    {            
        etVector l_faceNormal = OCT_TREE_BOX_NORMALS[l_sideIndex];
        PolySurf::Plane l_sidePlane(
            l_faceNormal,
            a_boxCenter + l_faceNormal.scale(fabs(l_halfBoxSize.dot(l_faceNormal))));
       
        for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
        {
            if (l_sidePlane.getIntersection(
                a_triangle->getVertexPosition(l_cornerIndex),
                a_triangle->getVertexPosition((l_cornerIndex + 1) % 3),
                l_intersectionDistance,
                l_intersection) )
            { 
                // if intersection is within the box bounds 
                // then return true
                etVector l_diff = l_intersection - a_boxCenter;                    

                if ( (l_faceNormal.x() != 0 || fabs(l_diff.x()) <= l_halfBoxSize.x()) &&
                     (l_faceNormal.y() != 0 || fabs(l_diff.y()) <= l_halfBoxSize.y()) &&
                     (l_faceNormal.z() != 0 || fabs(l_diff.z()) <= l_halfBoxSize.z()) )
                {
                    return true;
                }
            }
        } // triangle edges
    } // sides
    
    // Find the position of the box corners
    PolySurf::Plane l_triPlane(a_triangle->getNormal(), a_triangle->getVertexPosition(0));
    etVector l_boxCorners[8];

    for(int l_boxCornerIndex = 0; l_boxCornerIndex < 8; l_boxCornerIndex++)
    {
        l_boxCorners[l_boxCornerIndex] = 
             a_boxCenter + etVector(
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].x() * l_halfBoxSize.x(),
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].y() * l_halfBoxSize.y(),
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].z() * l_halfBoxSize.z());
    }    

    // Check for intersection between the box edges 
    // and the triangle
    for(int l_edgeIndex = 0; l_edgeIndex < 24; l_edgeIndex += 2)
    {
        if (l_triPlane.getIntersection(
            l_boxCorners[OCT_TREE_BOX_EDGES[l_edgeIndex]],
            l_boxCorners[OCT_TREE_BOX_EDGES[l_edgeIndex+1]],
            l_intersectionDistance,
            l_intersection) )
        {
            if (a_triangle->contains(l_intersection))
            {                    
                return true;
            }
        }
    }

    return false;
 }


//
// Collider member functions
//
etOctTree::Collider::Collider(const etVector& a_start, 
                            const etVector& a_end,
                            TriangleTest& a_triangleTest)
    :m_currentTriangle(0),
     m_start(a_start),
     m_end(a_end),
     m_largestDistanceRatio(0),
     m_triangleTest(a_triangleTest)
{
    // Do nothing
}

bool etOctTree::Collider::collide(const item_list_type& a_triangles)
{
    bool l_isNewCollison = false;
    for(int l_index = 0; l_index < a_triangles.size(); l_index++)
    {
        PolySurf::Plane l_plane(a_triangles[l_index]->getNormal(),
            a_triangles[l_index]->getVertexPosition(0));  
        etVector l_intersectPoint;
        double l_distanceRatio;
        if (l_plane.getIntersection(m_end, m_start, l_distanceRatio, l_intersectPoint))
        {
            // Check for largest distance ratio to find the point furthest
            // from end (ie closest to start)
            if (l_distanceRatio > m_largestDistanceRatio &&
                a_triangles[l_index]->contains(l_intersectPoint) &&
                 m_triangleTest.operator()(a_triangles[l_index]))
            {
                m_currentTriangle = a_triangles[l_index];
                m_largestDistanceRatio = l_distanceRatio;
                l_isNewCollison = true;
            }                  
        }
    }
    
    return l_isNewCollison;
}

void etOctTree::ParentNode::collide(Collider& a_collider, const etVector& a_nodeCenter, const etVector& a_nodeSize)
{            
    etVector l_childSize = a_nodeSize.scale(0.5);
    etVector l_childOffset = l_childSize.scale(0.5);

    etVector& l_start = a_collider.getStart();
    etVector& l_end = a_collider.getEnd();
    
    std::pair<etVector, etVector> l_line(l_start, l_end);

    // order octants          
    for(int x = 0; x < 2; x++)
    {
        for(int y = 0; y < 2; y++)
        {
            for(int z = 0; z < 2; z++)
            {
                int l_childIndex = 0;
                if ((x > 0) == (l_start.x() < l_end.x()))
                {
                    l_childIndex += 4; //XXX constant 
                }
                if ((y > 0) == (l_start.y() < l_end.y()))
                {
                    l_childIndex += 2;
                }
                if ((z > 0) == (l_start.z() < l_end.z()))
                {
                    l_childIndex += 1;
                }
                
                etVector l_childCenter = a_nodeCenter + etVector(
                        OCT_TREE_BOX_CORNERS[l_childIndex].x() * l_childOffset.x(),
                        OCT_TREE_BOX_CORNERS[l_childIndex].y() * l_childOffset.y(),
                        OCT_TREE_BOX_CORNERS[l_childIndex].z() * l_childOffset.z());
                        
                if (isLineInBox(l_line, l_childCenter, l_childSize))
                {
                    if (m_children[l_childIndex])
                    {
                        m_children[l_childIndex]->collide(a_collider, l_childCenter, l_childSize);
                    }

                    // Return if a triangle has been found
                    if (a_collider.getTriangle() != 0)
                    {
                        return;
                    }
                }
            }
        }
    }
}
 
void etOctTree::ParentNode::forEachItem(ItemListOperation& a_itemListTest)
{
    for(int l_childIndex = 0; l_childIndex < 8; l_childIndex++)
    {
        if (m_children[l_childIndex] != 0)
        {
            m_children[l_childIndex]->forEachItem(a_itemListTest);
        }
    }
}

bool etOctTree::Node::contains(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const
{
    etVector l_diff = a_targetPos - a_nodeCenter;
    etVector l_halfSize = a_nodeSize.scale(0.5);
   
    return (fabs(l_diff.x()) < l_halfSize.x() &&
            fabs(l_diff.y()) < l_halfSize.y() &&
            fabs(l_diff.z()) < l_halfSize.z());  
}

const etOctTree::Node* etOctTree::LeafNode::getNodeContaining(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const
{
    if (contains( a_targetPos, a_nodeCenter, a_nodeSize))
    {
        return this;
    }

    return 0;
}

const etOctTree::Node* etOctTree::ParentNode::getNodeContaining(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const
{    
    etVector l_childSize = a_nodeSize.scale(0.5);
    etVector l_childOffset = l_childSize.scale(0.5);

    // Check that the point is in bounds
    if (contains( a_targetPos, a_nodeCenter, a_nodeSize))
    {
        for(int l_childIndex = 0; l_childIndex < 8; l_childIndex++)
        {
            Node* l_currentChild = m_children[l_childIndex];

            if (l_currentChild)
            {
                // Find the center of the current child
                etVector l_childCenter = a_nodeCenter + etVector(
                    OCT_TREE_BOX_CORNERS[l_childIndex].x() * l_childOffset.x(),
                    OCT_TREE_BOX_CORNERS[l_childIndex].y() * l_childOffset.y(),
                    OCT_TREE_BOX_CORNERS[l_childIndex].z() * l_childOffset.z());
                
                const Node* l_childNode = l_currentChild->getNodeContaining(a_targetPos, l_childCenter, l_childSize);
                if (l_childNode != 0)
                {
                    return l_childNode;
                }
            }
        }
    }

    return 0;
}


etOctTree::Node* etOctTree::ParentNode::subdivide(const etVector& a_nodeCenter, const etVector& a_nodeSize, int a_depth, int& a_maxDepth)
{
    int l_childDepth = a_depth+1;
    etVector l_childSize = a_nodeSize.scale(0.5);
    etVector l_childOffset = l_childSize.scale(0.5);

    for(int l_childIndex = 0; l_childIndex < 8; l_childIndex++)
    {
        Node* l_currentChild = m_children[l_childIndex];

        if (l_currentChild)
        {
            // Find the center of the current child
            etVector l_childCenter = a_nodeCenter + etVector(
                OCT_TREE_BOX_CORNERS[l_childIndex].x() * l_childOffset.x(),
                OCT_TREE_BOX_CORNERS[l_childIndex].y() * l_childOffset.y(),
                OCT_TREE_BOX_CORNERS[l_childIndex].z() * l_childOffset.z());
                
            Node* l_dividedNode = l_currentChild->subdivide(l_childCenter, l_childSize, l_childDepth, a_maxDepth);

            // If the child was subdivided then delete the 
            // old pointer and subdivide the child again to
            // ensure it's children are subdivded
            //
            // This also means that the child changed from a leaf
            // to a parent node, so it is safe to delte the old leaf node
            if (l_currentChild != l_dividedNode)
            {
                m_children[l_childIndex] = l_dividedNode;
                delete l_currentChild;

                m_children[l_childIndex]->subdivide(l_childCenter, l_childSize, l_childDepth, a_maxDepth);
            }
        }
    }

    return this;
}

void etOctTree::Node::draw(const etVector& a_nodeCenter, const etVector& a_nodeSize) const
{    
    etVector l_boxCorners[8];
    etVector l_halfNodeSize = a_nodeSize.scale(0.5);
    for(int l_boxCornerIndex = 0; l_boxCornerIndex < 8; l_boxCornerIndex++)
    {
        l_boxCorners[l_boxCornerIndex] = 
             a_nodeCenter + etVector(
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].x() * l_halfNodeSize.x(),
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].y() * l_halfNodeSize.y(),
                OCT_TREE_BOX_CORNERS[l_boxCornerIndex].z() * l_halfNodeSize.z());
    } 
    
    glBegin(GL_LINES);   
    for(int l_edgeIndex = 0; l_edgeIndex < 24; l_edgeIndex += 2)
    {
        
        etVector& l_point0 =  l_boxCorners[OCT_TREE_BOX_EDGES[l_edgeIndex]];
        etVector& l_point1 =  l_boxCorners[OCT_TREE_BOX_EDGES[l_edgeIndex+1]];
      
        glVertex3d(l_point0.x(), l_point0.y(), l_point0.z());
        glVertex3d(l_point1.x(), l_point1.y(), l_point1.z());

    }  
    glEnd();
}

void etOctTree::ParentNode::draw(const etVector& a_nodeCenter, const etVector& a_nodeSize) const
{
    Node::draw(a_nodeCenter, a_nodeSize);
    etVector l_childSize = a_nodeSize.scale(0.5);
    etVector l_childOffset = l_childSize.scale(0.5);

    for(int l_octantIndex = 0; l_octantIndex < 8; l_octantIndex++)
    {
        if (m_children[l_octantIndex] != 0)
        {    
            etVector l_childCenter = a_nodeCenter + etVector(
                    OCT_TREE_BOX_CORNERS[l_octantIndex].x() * l_childOffset.x(),
                    OCT_TREE_BOX_CORNERS[l_octantIndex].y() * l_childOffset.y(),
                    OCT_TREE_BOX_CORNERS[l_octantIndex].z() * l_childOffset.z());

            m_children[l_octantIndex]->draw(l_childCenter, l_childSize);
        }
    }
}

void etOctTree::LeafNode::collide(Collider& a_collider, const etVector& a_nodeCenter, const etVector& a_nodeSize)
{            
    // Do the collision 
    a_collider.collide(m_itemList);
}        
       
etOctTree::Node* etOctTree::LeafNode::subdivide(const etVector& a_nodeCenter, const etVector& a_nodeSize, int a_depth, int& a_maxDepth)
{                             
    bool l_isCrossingOctant[8];
    bool l_hasChildren = false;
    LeafNode* l_children[8] = {0,0,0,0,0,0,0,0};

    // Check if node should be subdivided
    //XXX Should generalise the test for subdivide depth
    if (a_depth >= OCT_TREE_MAX_DEPTH
        ||  m_itemList.size() <= OCT_TREE_MAX_TRIANGLES_PER_NODE)                
    {
        return this;
    }

    int l_octIndex;
    // Loop through all triangles
    for(int l_index = 0; l_index < m_itemList.size(); l_index++)
    {                
        for(l_octIndex = 0; l_octIndex < 8; l_octIndex++)               
        {
            l_isCrossingOctant[l_octIndex] = false;
        }

        item_type l_tri = m_itemList[l_index];
               
        //check for octant, (x,y,z > halfway)
        bool l_crossCheck[] = 
        {
            l_tri->getVertexPosition(0).x() > a_nodeCenter.x(),
            l_tri->getVertexPosition(0).y() > a_nodeCenter.y(),
            l_tri->getVertexPosition(0).z() > a_nodeCenter.z(),

            l_tri->getVertexPosition(1).x() > a_nodeCenter.x(),
            l_tri->getVertexPosition(1).y() > a_nodeCenter.y(),
            l_tri->getVertexPosition(1).z() > a_nodeCenter.z(),

            l_tri->getVertexPosition(2).x() > a_nodeCenter.x(),
            l_tri->getVertexPosition(2).y() > a_nodeCenter.y(),
            l_tri->getVertexPosition(2).z() > a_nodeCenter.z()
        };

        // Check if all vertices are in one octant                
        bool l_isXCenterCrossed = (l_crossCheck[0] != l_crossCheck[3] || l_crossCheck[0] != l_crossCheck[6]);
        bool l_isYCenterCrossed = (l_crossCheck[1] != l_crossCheck[4] || l_crossCheck[1] != l_crossCheck[7]);
        bool l_isZCenterCrossed = (l_crossCheck[2] != l_crossCheck[5] || l_crossCheck[2] != l_crossCheck[8]);
           
        if (!l_isXCenterCrossed &&
            !l_isYCenterCrossed &&
            !l_isZCenterCrossed)
        {                   
            // Get the octant which contains all the vertices
            int l_octantIndex = l_crossCheck[0] ? 4 : 0;
            l_octantIndex += l_crossCheck[1] ? 2 : 0;
            l_octantIndex += l_crossCheck[2] ? 1 : 0;
            l_isCrossingOctant[l_octantIndex] = true;
        }
        // Else find all crossed octants
        else
        {                 
            // Check for valid octects          
            etVector l_octantSize = a_nodeSize.scale(0.5);
            etVector l_octantOffset = l_octantSize.scale(0.5);
            for(int l_octantIndex = 0; l_octantIndex < 8; l_octantIndex++)
            {                       
                 bool l_xCheck = 
                    // X crossed over
                    l_isXCenterCrossed || 
                     // Right side index and all vertices are on right side
                    ((l_octantIndex & OCTANT_X_MASK) && (l_crossCheck[0])) ||
                    // Left side index and all vertices are on left side
                    (!(l_octantIndex & OCTANT_X_MASK) && !l_crossCheck[0]);

                 bool l_yCheck = 
                    l_xCheck && 
                    (l_isYCenterCrossed ||                        
                    ((l_octantIndex & OCTANT_Y_MASK) && (l_crossCheck[1])) ||
                    (!(l_octantIndex & OCTANT_Y_MASK) && !l_crossCheck[1]));

                 bool l_zCheck = 
                    l_yCheck && 
                    (l_isZCenterCrossed || 
                    ((l_octantIndex & OCTANT_Z_MASK) && (l_crossCheck[2])) ||
                    (!(l_octantIndex & OCTANT_Z_MASK) && !l_crossCheck[2]));

                 // If the triangle bounds overlap the octant
                 // then check for triangle/octant intersection
                 if (l_zCheck)
                 {                            
                     etVector l_octantCenter = 
                        a_nodeCenter + etVector(
                        OCT_TREE_BOX_CORNERS[l_octantIndex].x() * l_octantOffset.x(),
                        OCT_TREE_BOX_CORNERS[l_octantIndex].y() * l_octantOffset.y(),
                        OCT_TREE_BOX_CORNERS[l_octantIndex].z() * l_octantOffset.z());                    

                     if (isTriangleInBox(l_tri, l_octantCenter, l_octantSize))
                     {                         
                        l_isCrossingOctant[l_octantIndex] = true;
                     }
                 }
            }                                 
        }

        // Add the current triangle to all overlapped octects
        for(int l_octantIndex = 0; l_octantIndex < 8; l_octantIndex++)
        {
            if (l_isCrossingOctant[l_octantIndex])
            {                       
                l_hasChildren = true;
                if (l_children[l_octantIndex] == 0)
                {   
                    l_children[l_octantIndex] = new LeafNode();
                }
                
                l_children[l_octantIndex]->getItemList().push_back(l_tri);
            }
        }
    } // Triangle loop                                                        

    // If all the children contain the same number of triangles
    // as this node, then don't subdivide
    // XXX may not want this
    for(l_octIndex = 0; l_octIndex < 8; l_octIndex++)
    {
        if (l_children[l_octIndex] != 0 &&
            l_children[l_octIndex]->getItemList().size() != m_itemList.size() )
        {
            break;
        }
    }

    if (l_octIndex == 8)
    {
        l_hasChildren = false;
        for(l_octIndex = 0; l_octIndex < 8; l_octIndex++)
        {
            delete l_children[l_octIndex];
        }
    }

    if (l_hasChildren)
    {
        ParentNode* l_newNode = new ParentNode();                

        for(l_octIndex = 0; l_octIndex < 8; l_octIndex++)
        {
            l_newNode->setChild(l_octIndex, l_children[l_octIndex]);
        }

        if (a_depth+1 > a_maxDepth)
        {
            a_maxDepth = a_depth+1;
        }
        return l_newNode;
    }
    
    return this;
}

//
// OctTree member functions  
//

void etOctTree::addTriangles(const item_list_type& a_triangles)
{
    //get bounds and add all triangles to topNode
    etVector l_minBound = a_triangles[0]->getVertexPosition(0);
    etVector l_maxBound = a_triangles[0]->getVertexPosition(0);
    LeafNode* l_topNode = new LeafNode();

    for(int l_triIndex = 0;  l_triIndex < a_triangles.size(); l_triIndex++)
    {
        l_topNode->getItemList().push_back(a_triangles[l_triIndex]);
        for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
        {
            etVector& l_vertexPosition = a_triangles[l_triIndex]->getVertexPosition(l_cornerIndex);
            l_minBound = etVector(               
                l_vertexPosition.x() < l_minBound.x() ? l_vertexPosition.x() : l_minBound.x(),
                l_vertexPosition.y() < l_minBound.y() ? l_vertexPosition.y() : l_minBound.y(),
                l_vertexPosition.z() < l_minBound.z() ? l_vertexPosition.z() : l_minBound.z());

            l_maxBound = etVector(               
                l_vertexPosition.x() > l_maxBound.x() ? l_vertexPosition.x() : l_maxBound.x(),
                l_vertexPosition.y() > l_maxBound.y() ? l_vertexPosition.y() : l_maxBound.y(),
                l_vertexPosition.z() > l_maxBound.z() ? l_vertexPosition.z() : l_maxBound.z());
        }
    }

    // subdivide cells until limit reached    
    m_topNodeCenter = (l_minBound + l_maxBound).scale(0.5);
    m_topNodeSize = (l_maxBound - l_minBound).scale(OCTTREE_BOUNDS_ALLOWANCE);

    int l_maxDepth = 0;
    
    m_topNode = l_topNode->subdivide(m_topNodeCenter, m_topNodeSize,0,l_maxDepth);
    
    // If a new node was created then the topnode was subdivided
    //  and is now a parent node, subdivide again to get the nodes
    //  children.    
    if (l_topNode != m_topNode)
    {
        delete l_topNode;
        m_topNode->subdivide(m_topNodeCenter, m_topNodeSize,0,l_maxDepth);
    }
}
 
etOctTree::item_type etOctTree::getCollision(const etVector& a_start, const etVector& a_end) const
{
    // Finds the first triangle which the line intersects
    TriangleTest a_triTest; // Default triangletest always returns true
    Collider l_collider(a_start, a_end, a_triTest);
    m_topNode->collide(l_collider, m_topNodeCenter, m_topNodeSize);
    return l_collider.getTriangle();        
}

// Returns the first intersected triangle which passes triangle test
etOctTree::item_type etOctTree::getCollisionIf(const etVector& a_start, const etVector& a_end, TriangleTest& a_triangleTest) const
{
    Collider l_collider(a_start, a_end, a_triangleTest);
    m_topNode->collide(l_collider, m_topNodeCenter, m_topNodeSize);
    return l_collider.getTriangle(); 
}
