// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// user.h -- contains the User class.
//
// $Revision: #2 $
// $Change: 1931 $
//

#ifndef POLYSURF_OCT_TREE_H
#define POLYSURF_OCT_TREE_H

// Includes
#include <et/vector.h>
#include "triangle.h"//<et/objects/triangle.h>
#include "vector"
#include <functional>

//
// Class:
//     etOctTree
//
// Description:
//     A spatial partitioning structure used to allow
//      efficient culling of triangles during collision detection.
//
// Exceptions:
//     None
//
// Notes:
//     None
//
// See Also Classes:
//     None
//
// See Also Functions:
//     None
//
class etOctTree
{
    public:

        typedef PolySurf::Triangle* item_type;
        typedef std::vector<item_type> item_list_type;

        // The index of each octant within a parent node
        // octant index = x * 4 + y*2 + z
        enum OctantIndexType
        {
            OIT_LBB, // Left,  Bottom,  Back octant
            OIT_LBF, // Left,  Bottom,  Front
            OIT_LTB, // Left,  Top,     Back
            OIT_LTF, // Left,  Top,     Front
            OIT_RBB, // Right, Bottom,  Back
            OIT_RBF, // Right, Bottom,  Front
            OIT_RTB, // Right, Top,     Back
            OIT_RTF // Right, Top,     Front
        };
    
        // function object to allow additional requirements
        // for intersecting a triangle
        class TriangleTest : public std::unary_function<item_type, bool>
        {
            public:
                virtual operator()(item_type a_item)
                {
                    return true;
                }
        };
        
        // Base class for an operation to be called on the 
        // item_list of each leaf node.
        class ItemListOperation : public std::unary_function<item_list_type&, bool>
        {
            public:
                virtual operator()(item_list_type& a_item)
                {
                    return true;
                }
        };
        
        // The collider tests lists oftriangles for 
        // an intersection with the stored line, abd
        // records the collison closest to the strat of the 
        // line.
        class Collider
        {
            public:

                // Constructor, takes the end points of the 
                // line use in collision tests.
                // triangleTest can be used to 
                // set additional requirements for colliding with
                // a triangle.                
                Collider(const etVector& a_start,
                         const etVector& a_end,         
                         TriangleTest& a_triangleTest);

                // Test for collison with all triangle in a_triangles
                bool collide(const item_list_type& a_triangles);

                // Return the intersected triangle closest to start of the line.
                const item_type getTriangle() const
                {
                    return m_currentTriangle;
                }
          
                // Return the start of the line.
                etVector getStart() const
                {
                    return m_start;
                }

                // Return the start of the line.
                etVector getEnd() const
                {
                    return m_end;
                }

            private:
           
                // Additional test for intersecting triangles.
                TriangleTest& m_triangleTest;

                // Closest intersected triangle.
                item_type m_currentTriangle;

                // Distance ratio to the line end for the closest triangle
                double m_largestDistanceRatio;
               
                // Line start.
                etVector m_start;

                // Line end.
                etVector m_end;
        };

        //
        // Class:
        //     etOctTree
        //
        // Description:
        //     Base class for nodes in the oct-tree.
        //
        class Node
        {       
            public:

                // Returns the lowest level node which contains a_targetPos.
                virtual const Node* getNodeContaining(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const = 0;

                // Returns true if the current node contains a_targetPos
                bool contains(const etVector& a_targetPos, 
                             const etVector& a_nodeCenter, 
                             const etVector& a_nodeSize) const;

                // Draw an outline for this node
                virtual void draw(const etVector& a_nodeCenter, 
                                 const etVector& a_nodeSize) const;

                // Attempt collison with this nodse children
                virtual void collide(Collider& a_collider,
                             const etVector& a_nodeCenter,
                             const etVector& a_nodeSize) = 0;

                // 
                virtual Node* subdivide(const etVector& a_nodeCenter,
                                       const etVector& a_nodeSize,
                                       int a_depth,
                                       int& a_maxDepth) = 0;      

                // 
                virtual void forEachItem(ItemListOperation& a_itemListTest) = 0;
                
        };

        //
        // Class:
        //     ParentNode
        //
        // Description:
        //     Node which contains other nodes.
        //
        class ParentNode : public Node
        {
            public:

                // Defualt constructor
                ParentNode()
                {
                    for(int i = 0; i < 8; i++)
                    {
                        m_children[i] = 0;
                    }
                }

                // Destructor.
                ~ParentNode()
                {
                    for(int i = 0; i < 8; i++)
                    {
                        delete m_children[i];
                    }
                }

                // Returns the lowest level node which contains a_targetPos.               
                virtual const Node* getNodeContaining(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const;

                // Draw an outline for this node.
                virtual void draw(const etVector& a_nodeCenter, 
                                 const etVector& a_nodeSize) const;
                 
                // Collide with each of this nodes children.
                virtual void collide(Collider& a_collider,
                             const etVector& a_nodeCenter,
                             const etVector& a_nodeSize);

                // Subdivide all the nodes children.
                virtual Node* subdivide(const etVector& a_nodeCenter,
                                       const etVector& a_nodeSize,
                                       int a_depth,
                                       int& a_maxDepth);

                // Run a_itemListTest on the itemlist of all leaf nodes
                // below this node.
                virtual void forEachItem(ItemListOperation& a_itemListTest);                
                
                // Set a child node.
                inline void setChild(int l_childIndex, Node* a_node)
                {
                    assert(l_childIndex >= 0 && l_childIndex < 8);
                    m_children[l_childIndex] = a_node;
                }
                  
            private:
                Node* m_children[8];
        };
  
        //
        // Class:
        //     LeafNode
        //
        // Description:
        //     Bottom level node which contains triangles.
        //
        class LeafNode : public Node
        {  
        
            public:

                // Returns this node if it contains a_targetPos, else return 0
                virtual const Node* getNodeContaining(const etVector& a_targetPos, const etVector& a_nodeCenter, const etVector& a_nodeSize) const;

                // Collide with all contained triangles.
                virtual void collide(Collider& a_collider,
                                    const etVector& a_nodeCenter, 
                                    const etVector& a_nodeSize);       
        
                // Subdivide the node.
                // Returns a new parent node if subdividion succeds.
                virtual Node* subdivide(const etVector& a_nodeCenter,
                                        const etVector& a_nodeSize,
                                        int a_depth,
                                        int& a_maxDepth);    
                
                // Call itemListTest on this nodes triangle list.
                virtual void forEachItem(ItemListOperation& a_itemListTest)
                {
                    a_itemListTest.operator()(m_itemList);
                }

                // Return the list of triangles contained in the node.
                virtual item_list_type& getItemList()
                {
                    return m_itemList;
                }

                // Return the list of triangles contained in the node.
                virtual const item_list_type& getItemList() const
                {
                    return m_itemList;
                }
                
            private:

                item_list_type m_itemList;
        };  

    public:

        // Default constructor.
        etOctTree():m_topNode(0)
        {
            // Do nothing
        }

        // Destructor.
        ~etOctTree()
        {
            delete m_topNode;
        }
      
        // Initiliase the contained triangles of the oct-tree
        // and set it's top level bounds.
        void addTriangles(const item_list_type& a_triangles);       
         
        // Get the triangle which is closest to start and 
        // intersected by the line start to end.
        item_type getCollision(const etVector& a_start,
                                const etVector& a_end) const;
        
        // Get the triangle which is closest to start and 
        // intersected by the line start to end and passes
        // triangle test.
        item_type getCollisionIf(const etVector& a_start,
                                 const etVector& a_end,
                                 TriangleTest& a_triangleTest) const;

        // Draw the oct-tree.
        void draw() const
        {
            if (m_topNode)
            {
                m_topNode->draw(m_topNodeCenter, m_topNodeSize);
            }
        }

        // Return the node which cotains point.
        const Node* getNodeContainingPoint(const etVector& a_point) const
        {
            const Node* l_result = 0;
            if (m_topNode)
            {
                l_result = m_topNode->getNodeContaining(a_point, m_topNodeCenter, m_topNodeSize);
            }

            return l_result;
        }

        // Returns the center of the oct-trees bounding box.
        etVector getCenter() const
        {
            return m_topNodeCenter;
        }

        // Returns the size of the oct-trees bounding box.
        etVector getSize() const
        {
            return m_topNodeSize;
        }

        // Call a_itemListTest for the item list of each leaf node.
        void forEachItem(ItemListOperation& a_itemListTest)
        {
            m_topNode->forEachItem(a_itemListTest);            
        }

    private:
      
        // Not implemented.
        etOctTree(const etOctTree& other);        

        // Not implemented.
        etOctTree& operator=(const etOctTree& other);

        Node* m_topNode;
     
        etVector m_topNodeCenter;
        etVector m_topNodeSize;
};

// Utility functions

// Returns true if the triangle overlaps the box
bool isTriangleInBox(etOctTree::item_type a_triangle, const etVector& a_boxCenter, const etVector& a_boxSize);

// Check for intersection between a box and a line
bool isLineInBox(std::pair<etVector, etVector> a_line,
                 const etVector& a_boxCenter,
                 const etVector& a_boxSize);

#endif
