#include "trisurface.h"

#include "cssysdef.h"
#include "iengine/mesh.h"
#include "igeom/polymesh.h"
#include "igeom/objmodel.h"
#include "iengine/movable.h"
#include "imesh/object.h"
#include "iutil/object.h"

#include "phantom.h" 

CS_IMPLEMENT_PLUGIN

SCF_IMPLEMENT_IBASE (csPhantom)
	SCF_IMPLEMENTS_INTERFACE (iPhantom)
	SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iComponent)
SCF_IMPLEMENT_IBASE_END

SCF_IMPLEMENT_EMBEDDED_IBASE (csPhantom::Component)
  SCF_IMPLEMENTS_INTERFACE (iComponent)
SCF_IMPLEMENT_EMBEDDED_IBASE_END


/** Constructor. Runs initialize. */
csPhantom::csPhantom (iBase *iParent) {
	SCF_CONSTRUCT_IBASE (iParent);
	SCF_CONSTRUCT_EMBEDDED_IBASE (scfiComponent);
	world = NULL;
	scene = NULL;
	phantomExists = false;
	initialize();
}

/** Destructor. */
csPhantom::~csPhantom() {
	if (world) delete world;
	if (scene) delete scene;
}

/** Initializes the GHOST scene and creates the haptic world. */
void csPhantom::initialize(){
	createWorld();
	createScene();
}

/** Returns true if the phantom is correctly initialized. */
bool csPhantom::getValidConstruction() const{
	return phantomExists;
}

/** Implements the iComponent interface. */
bool csPhantom::Initialize (iObjectRegistry* iobject_reg)
{
  object_reg = iobject_reg;
  return true;
}


/** Starts the servo loop. */
void csPhantom::start() const{
	// Start the servo loop & wait until it is running
	if (phantomExists) {
		if (scene->startServoLoop() == 0)
		{
			::exit(-1);
		}
		while (scene->getDoneServoLoop() == TRUE)
		{
			//XXX replace system specific operation
			::Sleep(1);
		}
	}
	else {
		printf("Error: Cannot start the plugin because no phantom device exists.\n");
	}	
}

/** Stops the servo loop. */
void csPhantom::stop() const{
	// Stop the servo loop & wait until it is stopped
	if (phantomExists) {
		scene->stopServoLoop();
		while (scene->getDoneServoLoop() == FALSE)
		{
			//XXX replace system specific operation
			::Sleep(1);
		}
	}
	else {
		printf("Error: Cannot stop the plugin because no phantom device exists.\n");
	}
}

/** Creates the GHOST scene. */ 
void csPhantom::createScene(){	
	scene = new gstScene;
	
	// Create the GHOST scene graph root
    root = new gstSeparator;
    scene->setRoot(root);

    // Add a transform for the phantom
	phantomTransform = new gstSeparator;
    root->addChild(phantomTransform);

	// Add a force field to extract phantom information & to return
    // forces and torques
    gstForceField* forceField = new ForceField(world);
    forceField->boundByBox(gstPoint(0.0,0.0,0.0),
		5000.0,
		5000.0,
		5000.0);
    forceField->setAttenuationDistance(5.0);
    root->addChild(forceField);
	
	// Turn off the quit on device fault flag
    scene->setQuitOnDevFault(false);
	
	// Create a new phantom & add it to the GHOST scene graph
	phantom = new gstPHANToM("Default PHANToM", true);
	if (phantom->getValidConstruction())
		phantomExists = true;
	phantomTransform->addChild(phantom);
}

/** Creates a new world object. */
void csPhantom::createWorld(){
	world = new World();
}

/** Calls the setMeshList function of the world object. */
void csPhantom::setMeshList(const iMeshList* meshList){
	if (phantomExists)
		world->setMeshList(meshList);
}

/** Sets the position of the PHANTOM workspace to originate at x,y,z. */
void csPhantom::setWorkspacePosition(double x, double y, double z) {
	phantomTransform->setPosition(x*POS_ETOUCH_TO_GHOST, y*POS_ETOUCH_TO_GHOST, -z*POS_ETOUCH_TO_GHOST);
}

/** Rotates the PHANTOM workspace with an amount of a radians around the vector given by x,y,z. */
void csPhantom::rotateWorkspace(double a, double x, double y, double z) {
	phantomTransform->rotateRM(gstVector(-x,-y,-z),a);
}

/** Returns a vector with the current PHANTOM stylus position. */
csVector3 csPhantom::getStylusPosition() const{
	gstPoint gstPosition;
	phantom->getPosition_WC(gstPosition);
	return csVector3(gstPosition.x()*POS_GHOST_TO_ETOUCH, gstPosition.y()*POS_GHOST_TO_ETOUCH, -gstPosition.z()*POS_GHOST_TO_ETOUCH);
}

/** Returns a vector with the current PHANTOM gimbal angles. In radians. */
csVector3 csPhantom::getGimbalAngles() const {
	gstVector angles;
	angles = phantom->getGimbalAngles();
	return csVector3(angles.x(),angles.y(),angles.z());
}

/** Returns true if the stylus switch is pressed. */
bool csPhantom::getStylusSwitch() const{
	return phantom->getStylusSwitch();
}

/** Returns a vector with the current force. */
csVector3 csPhantom::getCurrentForce() const{
	gstVector gstForce = phantom->getReactionForce_WC();
	return csVector3(-(5.0*gstForce.x()), -(5.0*gstForce.y()), 5.0*gstForce.z());
}

/** Set forces on or off. */
void csPhantom::setForcesOn(bool forcesOn) {
	world->setForcesOn(forcesOn);
}

/** Returns a list containing pairs with indices to dynamic objects and their respective forces. */
std::pair<int,csVector3> csPhantom::getDynamicForcePair() const {
	return world->getDynamicForcePair();
}

/** Returns a list containing pairs with indices to dynamic objects and their respective forces. */
std::list<std::pair<int,csVector3> > csPhantom::getDynamicForceList() const {
	return world->getDynamicForceList();
}

/** Returns true if the PHANTOM touches an object. */
bool csPhantom::isCollision() const{
	return world->collision;
}

SCF_IMPLEMENT_FACTORY (csPhantom)

SCF_EXPORT_CLASS_TABLE (phantom)
SCF_EXPORT_CLASS (csPhantom, "crystalspace.plugins.phantom", "phantom plugin")
SCF_EXPORT_CLASS_TABLE_END