#include "csutil/scf.h"
#include "iutil/comp.h"

#include <gstForceField.h>
#include <gstScene.h>
#include <gstSeparator.h>
#include <gstPHANToM.h>
#include <gstPoint.h>

#include "world.h"

struct iObjectRegistry;

SCF_VERSION (iPhantom, 0, 0, 1);

static const double POS_GHOST_TO_ETOUCH = 0.050;
static const double POS_ETOUCH_TO_GHOST = 1/0.050;
static const double FORCE_SCALE = 0.04;

struct iPhantom : public iBase
{
public:
	virtual void start() const = 0;
	virtual void stop() const = 0;
	virtual void setMeshList(const iMeshList* meshList) = 0;
	virtual void setWorkspacePosition(double x, double y, double z) = 0;
	virtual void rotateWorkspace(double a, double x, double y, double z) = 0;
	virtual void setForcesOn(bool forcesOn) = 0;
	virtual bool getValidConstruction() const = 0;
	virtual bool isCollision() const = 0;
	virtual bool getStylusSwitch() const = 0;
	virtual csVector3 getGimbalAngles() const = 0;
	virtual csVector3 getStylusPosition() const = 0;
	virtual csVector3 getCurrentForce() const = 0;
	virtual std::pair<int,csVector3> getDynamicForcePair() const = 0;
	virtual std::list<std::pair<int,csVector3> > getDynamicForceList() const = 0;
};



class csPhantom : public iPhantom {

private:
	bool phantomExists;
	gstScene* scene;
	gstSeparator* root;
	gstSeparator* phantomTransform;
	gstPHANToM* phantom;	
	gstBoolean stylusSwitch;
	gstPoint position;
	gstTransformMatrix rotationMatrix;
	World* world;
	iObjectRegistry* object_reg;
	
public:
	SCF_DECLARE_IBASE;
	
	

	csPhantom (iBase *iParent);
	~csPhantom();
	void start() const;
	void stop() const;
	void setMeshList(const iMeshList* meshList);
	void setWorkspacePosition(double x, double y, double z);
	void rotateWorkspace(double a, double x, double y, double z);
	void setForcesOn(bool forcesOn);
	bool getStylusSwitch() const;
	bool getValidConstruction() const;
	bool isCollision() const;
	csVector3 getStylusPosition() const;
	csVector3 getCurrentForce() const;
	csVector3 getGimbalAngles() const;
	bool Initialize (iObjectRegistry* iobject_reg);
	std::pair<int,csVector3> getDynamicForcePair() const;
	std::list<std::pair<int,csVector3> > getDynamicForceList() const;
	
	struct Component : public iComponent
	{
		SCF_DECLARE_EMBEDDED_IBASE (csPhantom);
		virtual bool Initialize (iObjectRegistry* object_reg)
		{
			return scfParent->Initialize (object_reg);
		}
	} scfiComponent;


private:
	void createScene();
	void createWorld();
	void initialize();
};

class ForceField : public gstForceField {
public:
	
	/** Constructor. */
	ForceField(World* w) : gstForceField() {
		world = w;
	}
	
	/** See gstForceField. Calculates and scales the force to the PHANTOM. */
	gstVector calculateForceFieldForce(gstPHANToM* a_phantom) {
		gstPoint phantomPos = a_phantom->getPosition_WC();
		phantomPos.setx(phantomPos.x() * POS_GHOST_TO_ETOUCH);
		phantomPos.sety(phantomPos.y() * POS_GHOST_TO_ETOUCH);
		phantomPos.setz(phantomPos.z() * POS_GHOST_TO_ETOUCH);
		etVector phantomVec(phantomPos.x(), phantomPos.y(), phantomPos.z());
		etVector etForce = world->calculateForces(phantomVec);	
		gstVector gstForce = gstVector(etForce.x(), etForce.y(), etForce.z());			
		gstForce *= FORCE_SCALE;

		if (etForce.length() != 0)
			world->collision = true;
		else
			world->collision = false;
		return gstForce;
	}
	
	/** See gstForceField. Sends the calculated force to the PHANTOM. */
	gstVector calculateForceFieldForce(gstPHANToM* a_phantom, gstVector& a_torques) {
		gstVector& force = calculateForceFieldForce(a_phantom);
		return force;
	}
	
private:
	World* world;
};
