// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// messages.cpp -- contains the Messages implementation.
//
// $Revision: #1 $
// $Change: 1787 $
//

// Includes
#include <et/common.h>
#include "orientation.h"//<et/orientation.h>

#include "triangle.h"//<et/objects/triangle.h>
#include "edgedata.h"//<et/objects/edge_data.h>

const double POLYSURF_TRIANGLE_EPSILON = 0.00000001;

PolySurf::Triangle::Triangle(Vertex* a_Vertex0, Vertex* a_Vertex1, Vertex* a_Vertex2)
{
    assert(a_Vertex0);
    assert(a_Vertex1);
    assert(a_Vertex2);

    m_vertices[0] = a_Vertex0;
    m_vertices[1] = a_Vertex1;
    m_vertices[2] = a_Vertex2;

    m_normal = (m_vertices[1]->m_position - m_vertices[0]->m_position).
        cross(m_vertices[2]->m_position - m_vertices[1]->m_position).
        normalize();
    
    //Triangle edges are set when the triangle is added
    m_edges[0] = 0;
    m_edges[1] = 0;
    m_edges[2] = 0;
    m_edges[3] = 0;
    m_edges[4] = 0;
    m_edges[5] = 0;  
}

PolySurf::Triangle* PolySurf::Triangle::getNeighbour(int a_edgeIndex, bool a_isFrontFace) const
{
    assert(a_edgeIndex >= 0 && a_edgeIndex < 3);
    if (!a_isFrontFace)
    {
        a_edgeIndex+= 3;
    }

    if (m_edges[a_edgeIndex] == 0)
    {
        return 0;
    }
    if (this == m_edges[a_edgeIndex]->getNeighbour0())
    {
        return m_edges[a_edgeIndex]->getNeighbour1();            
    }
    else if (this == m_edges[a_edgeIndex]->getNeighbour1())
    {
        return m_edges[a_edgeIndex]->getNeighbour0();           
    }

    return 0;
}

// Use angles to check for inside
bool PolySurf::Triangle::contains(const etVector& a_point) const
{
    double l_totalAngle = 0;
    for(int l_pointIndex = 0;
            l_pointIndex < 3;
            l_pointIndex++)
    {
        etVector l_line1 = getVertexPosition(l_pointIndex) - a_point;
        etVector l_line2 = getVertexPosition((l_pointIndex+1) % 3) - a_point;
        
        double l_length1 = l_line1.length();
        double l_length2 = l_line2.length();
        if (l_length1 * l_length2 < POLYSURF_TRIANGLE_EPSILON)
        {
            // Point is on a vertex
            return true;
        }        
        l_totalAngle += acos(l_line1.dot(l_line2) / (l_length2 * l_length1));
    }
    
    //Return true if angle == 2*pi
    return (fabs(l_totalAngle - etOrientation::PII*2) < POLYSURF_TRIANGLE_EPSILON); 
}

bool PolySurf::isTriangleOrdered(const PolySurf::Triangle* a_triangle,                        
                       const PolySurf::Vertex* a_edgePoint0, 
                       const PolySurf::Vertex* a_edgePoint1)
{
    bool l_result;
    for(int l_cornerIndex = 0; l_cornerIndex < 3; l_cornerIndex++)
    {
        if (a_triangle->getVertex(l_cornerIndex) == a_edgePoint0)
        {
            l_result =
                (a_triangle->getVertex((l_cornerIndex + 1) % 3) == a_edgePoint1);
            break;
        }
        else if (a_triangle->getVertex(l_cornerIndex) == a_edgePoint1)
        {
            l_result =
                (a_triangle->getVertex((l_cornerIndex + 2) % 3) == a_edgePoint0);
            break;
        }
    }

    // Raise an error if the triangle doesn't contain either point
    assert(l_cornerIndex != 3);

    return l_result;
}

 // Angle is positive in the normal direction of tri0.
double PolySurf::getAngleBetweenTriangles(const PolySurf::Triangle* a_tri0, 
                                const PolySurf::Triangle* a_tri1, 
                                const PolySurf::Vertex* a_edgePoint0, 
                                const PolySurf::Vertex* a_edgePoint1)
{
    if (a_tri0 == 0 || a_tri1 == 0)
    {
        return 0;
    }
    etVector l_normal0 = a_tri0->getNormal();
    etVector l_normal1 = a_tri1->getNormal();

    if (PolySurf::isTriangleOrdered(a_tri0, a_edgePoint0, a_edgePoint1) ==
        PolySurf::isTriangleOrdered(a_tri1, a_edgePoint0, a_edgePoint1))
    {    
        l_normal1 = l_normal1.scale(-1);
    }

    double l_dot = l_normal0.dot(l_normal1);

    double l_angle = etOrientation::PII - acos(l_dot);

    // Get the points from neighbours which aren't in this edge
    etVector l_otherPoint0;
    etVector l_otherPoint1;
    for(int l_cornerIndex=0; l_cornerIndex< 3; l_cornerIndex++)
    {
        if ((a_tri0->getVertex(l_cornerIndex) != a_edgePoint0) &&
            (a_tri0->getVertex(l_cornerIndex) != a_edgePoint1))
        {
            l_otherPoint0 = a_tri0->getVertexPosition(l_cornerIndex);
        }
        if ((a_tri1->getVertex(l_cornerIndex) != a_edgePoint0) &&
            (a_tri1->getVertex(l_cornerIndex) != a_edgePoint1))
        {
            l_otherPoint1 = a_tri1->getVertexPosition(l_cornerIndex);
        }
    }

    // Set the edge to be convex if the
    // neighbour0's normal points towards neighbour1.
    if ((l_otherPoint1 - l_otherPoint0).dot(l_normal0) < 0)
    {
        // Then normals point outwards, 
        // hence the edge is convex
        l_angle = 2 * etOrientation::PII - l_angle;
    }

    return l_angle;
}