// -*-c++-*-
//
// Copyright 2001 Novint Technologies, Inc. All rights reserved.
// Available only under license from Novint Technologies, Inc.
//
// triangle.h -- contains the triangle implementation.
//
// $Revision: #1 $
// $Change: 1787 $
//

#ifndef POLYSURF_TRIANGLE
#define POLYSURF_TRIANGLE

#include <et/vector.h>
#include <assert.h>

namespace PolySurf
{
    class EdgeData;

    //
    // Class:
    //     Vertex
    //
    // Description:
    //     A vertex composed of a point in space and a normal.
    //     Triangles have six edges recorded, 3 for the front 
    //     face and 3 for the rear face.    
    //
    // Exceptions:
    //     None
    //
    // Notes:
    //     None
    //
    // See Also Classes:
    //     Triangle
    //
    // See Also Functions:
    //     None
    //
    struct Vertex
    {
        // Constructor.
        Vertex(const etVector& a_position):
                m_position(a_position),
                m_normal(etVector())
        {
            // 
        }
        etVector m_position;
        etVector m_normal;
    };


    //
    // Class:
    //     etBox
    //
    // Description:
    //     A triangle, for use in a HapticTrisurface.    
    //
    // Exceptions:
    //     None
    //
    // Notes:
    //      The normal is calculated in the constructor,
    //      but is not updated by setVertex.
    //
    // See Also Classes:
    //     Vertex, EdgeData, etHapticTriSurface
    //
    // See Also Functions:
    //     None
    //
    class Triangle
    {           
        public:

            // Constructor, takes three vertices.
            Triangle(Vertex* a_Vertex0, Vertex* a_Vertex1, Vertex* a_Vertex2);

            // Change on the triangles vertices.
            // Doesn't update the normal.
            void setVertex(int l_index, Vertex* a_vertex)
            {
                assert(l_index >= 0 && l_index < 3);
                m_vertices[l_index] = a_vertex;
            }

            // Returns a vertex, from 0 to 2.
            inline Vertex* const getVertex(int l_index) const
            {
                assert(l_index > -1 && l_index < 3);
                return m_vertices[l_index];
            }

            // Returns the psotiion of the vertex.
            // Index is 0 to 2.
            inline etVector getVertexPosition(int l_index) const
            {                    
                return getVertex(l_index)->m_position;
            }

            // Returns the triangles normal.
            inline etVector getNormal() const
            {
                return m_normal;
            }

            // Return the edge data for the given edge and face of the triangle.
            inline const EdgeData* getEdgeData(int a_edgeIndex, bool a_isFrontFaceTouched) const
            {
                assert(a_edgeIndex > -1 && a_edgeIndex < 3);

                if (!a_isFrontFaceTouched)
                {
                    a_edgeIndex += 3;
                }

                return m_edges[a_edgeIndex];
            }

            // Sets the edge data for the given index,
            // the rear faces have index +3.
            inline setEdgeData(int a_edgeIndex,  EdgeData* a_edgeData)
            {
                assert(a_edgeIndex > -1 && a_edgeIndex < 6);
                m_edges[a_edgeIndex] = a_edgeData;
            }

            // Return the neighbour for a given egde and face.
            Triangle* getNeighbour(int a_edgeIndex, bool a_isFrontFace) const;
           
            // Return true if a_point is inside the triangle
            bool contains(const etVector& a_point) const;

        private:
            // Edges 0 to 2 are for the normal side
            // Edges 3 to 5 are for the anti-normal side.
            EdgeData* m_edges[6];
            Vertex* m_vertices[3];

            // Normalised normal
            etVector m_normal;
    };

    // Returns true if the a_edgePoint0 and a_edgePoint1 appear
    // in that order within the triangle.
    bool isTriangleOrdered(const Triangle* a_triangle,                        
                           const Vertex* a_edgePoint0, 
                           const Vertex* a_edgePoint1);

    // Retunrs the angle between two triangles.
    // Angle is positive in the normal direction of tri0.
    double getAngleBetweenTriangles(const Triangle* a_tri0, 
                                    const Triangle* a_tri1, 
                                    const Vertex* a_edgePoint0, 
                                    const Vertex* a_edgePoint1);

}

#endif