#include "cssysdef.h"
#include "iengine/mesh.h"
#include "igeom/polymesh.h"
#include "igeom/objmodel.h"
#include "iengine/movable.h"
#include "imesh/object.h"
#include "imesh/ball.h"
#include "iutil/object.h"
#include "imesh/thing/thing.h"
#include "imesh/sprite3d.h"
#include "ivaria/keyval.h"
#include "csutil/scanstr.h"

#include <utility>
#include <cstring>

#include <et/thread.h>

#include "world.h"

const int MAX_FORCE = 200; 

#undef GetObject;

/** Constructor. */
World::World(){	
	octree = NULL;
	collision = false;
	mutex = new etLock();
	forcesOn = true;
}

/** Destructor. */
World::~World(){
	if (octree) delete octree;
}

/** Returns the static objects close to the PHANTOM position by calling WorldOctree. */  
std::list<int> World::getNearbyStaticObjects(double x, double y, double z) const {
	return octree->getNearbyObjects(x,y,z);
}

/** Adds the list of static objects and their bounding boxes to the world. Sets the min and max boundaries of the WorldOctree. 
* Creates the WorldOctree. 
*/
void World::addStaticObjects(std::vector<etObject*> objects,std::vector<csBox3> boundBoxes,std::vector<double> minMax) {
	staticObjects = objects;
	boxes = boundBoxes;
	nbrOfStaticObjects = objects.size();
	octree = new WorldOctree(boundBoxes,minMax);
}

/** Adds the list of dynamic objects and their index list to the world. */
void World::addDynamicObjects(std::vector<etObject*> objects, std::vector<int> dynamicIndexVector) {
	dynamicObjects = objects;
	dynamicIndices = dynamicIndexVector; 
	nbrOfDynamicObjects = objects.size();
}

/** Turn on forces. */
void World::setForcesOn(bool on) {
	forcesOn = on;
}

/** Creates a haptic representation of all the mesh objects in csMeshList. Static objects are saved in diffrent vectors. Only adds objects
* that implements the iPolygonMesh interface.
*/
void World::setMeshList(const iMeshList* csMeshList) {
	printf("Creating haptic environment...\n");
	int polyCount;
	int vertexCount;
	int meshListSize = csMeshList->GetCount();
	int nbrOfTriangles;
	int meshIndex = 0;
	int nbrOfDynamicObjects = 0;
	std::list<etObject*> staticObjectList;
	std::list<etObject*> dynamicObjectList;
	std::list<csBox3> boundBoxes;
	std::list<int> dynamicIndexList;
	std::vector<double> minMaxVec(6);
	minMaxVec[0] = DBL_MAX;
	minMaxVec[1] = DBL_MIN;
	minMaxVec[2] = DBL_MAX;
	minMaxVec[3] = DBL_MIN;
	minMaxVec[4] = DBL_MAX;
	minMaxVec[5] = DBL_MIN;
	
	// Check all the meshes and create haptic trisurfaces.
	for(int i = 0;i<meshListSize;i++) {
		iMeshWrapper* meshWrapper = csMeshList->Get(i);
		iPolygonMesh* polyMesh = meshWrapper->GetMeshObject()->GetObjectModel()->GetPolygonMeshColldet();
		iObject* myObject = meshWrapper->QueryObject();

		// Just interrested in the ones that implements polymesh.
		if ((polyMesh != NULL) && (polyMesh->GetPolygonCount() != 0)) { 
			// Not the phantom cursor.
			if (strcmp(myObject->GetName(),"PHANTOM") != 0) {
				
				polyCount = polyMesh->GetPolygonCount();
				vertexCount = polyMesh->GetVertexCount();
				nbrOfTriangles = 0;
				std::vector<etVector> vertexVector(vertexCount);
				csMatrix3 matrix;
				matrix.m31 = -matrix.m31;matrix.m32 = -matrix.m32;matrix.m33 = -matrix.m33;
				int count = 1;

				// For all polygons. Divide the polygon into triangles.
				for(int j=0;j<polyCount;j++) {

					count = 1;

					// One triangle of the polygon.
					for(int k=0;k<polyMesh->GetPolygons()[j].num_vertices-2;k++) { 
						
						// First vertex.
						int a = polyMesh->GetPolygons()[j].vertices[0]; // Index in vertexVector.
						csVector3 csvertex1(polyMesh->GetVertices()[a].x,polyMesh->GetVertices()[a].y,polyMesh->GetVertices()[a].z);
						csvertex1 = operator *(matrix,csvertex1);   // Transformation from object to world				
						etVector vertex1(csvertex1.x,csvertex1.y,csvertex1.z);
						vertexVector[a] = vertex1;
						
						// Get the maximum and minimum world coordinates
						if(csvertex1.x < minMaxVec[0])
							minMaxVec[0] = csvertex1.x;
						
						if(csvertex1.x > minMaxVec[1])
							minMaxVec[1] = csvertex1.x;
						
						if(csvertex1.y < minMaxVec[2])
							minMaxVec[2] = csvertex1.y;
						
						if(csvertex1.y > minMaxVec[3])
							minMaxVec[3] = csvertex1.y;
						
						if(csvertex1.z < minMaxVec[4])
							minMaxVec[4] = csvertex1.z;
						
						if(csvertex1.z > minMaxVec[5])
							minMaxVec[5] = csvertex1.z;
						
						// Second vertex.
						a = polyMesh->GetPolygons()[j].vertices[count];
						csVector3 csvertex2(polyMesh->GetVertices()[a].x,polyMesh->GetVertices()[a].y,polyMesh->GetVertices()[a].z);
						csvertex2 = operator *(matrix,csvertex2);
						etVector vertex2(csvertex2.x,csvertex2.y,csvertex2.z);
						vertexVector[a] = vertex2;
						
						// Get the maximum and minimum world coordinates
						if(csvertex2.x < minMaxVec[0])
							minMaxVec[0] = csvertex2.x;
						
						if(csvertex2.x > minMaxVec[1])
							minMaxVec[1] = csvertex2.x;
						
						if(csvertex2.y < minMaxVec[2])
							minMaxVec[2] = csvertex2.y;
						
						if(csvertex2.y > minMaxVec[3])
							minMaxVec[3] = csvertex2.y;
						
						if(csvertex2.z < minMaxVec[4])
							minMaxVec[4] = csvertex2.z;
						
						if(csvertex2.z > minMaxVec[5])
							minMaxVec[5] = csvertex2.z;
						
						// Third vertex.
						a = polyMesh->GetPolygons()[j].vertices[count+1];
						csVector3 csvertex3(polyMesh->GetVertices()[a].x,polyMesh->GetVertices()[a].y,polyMesh->GetVertices()[a].z);
						csvertex3 = operator *(matrix,csvertex3);
						etVector vertex3(csvertex3.x,csvertex3.y,csvertex3.z);
						vertexVector[a] = vertex3;
						
						// Get the maximum and minimum world coordinates
						if(csvertex3.x < minMaxVec[0])
							minMaxVec[0] = csvertex3.x;
						
						if(csvertex3.x > minMaxVec[1])
							minMaxVec[1] = csvertex3.x;
						
						if(csvertex3.y < minMaxVec[2])
							minMaxVec[2] = csvertex3.y;
						
						if(csvertex3.y > minMaxVec[3])
							minMaxVec[3] = csvertex3.y;
						
						if(csvertex3.z < minMaxVec[4])
							minMaxVec[4] = csvertex3.z;
						
						if(csvertex3.z > minMaxVec[5])
							minMaxVec[5] = csvertex3.z;
						
						nbrOfTriangles++;
						count++;
					}	
				}
				
				// Create the correct coordinate vector
				int l = 0;
				std::vector<int> coordinateIndex(nbrOfTriangles*4);
				
				for(int m=0;m<polyCount;m++) {
					count = 1;
					
					for(int o=0;o<polyMesh->GetPolygons()[m].num_vertices-2;o++) {
						coordinateIndex[l] = polyMesh->GetPolygons()[m].vertices[0];
						coordinateIndex[l+1] = polyMesh->GetPolygons()[m].vertices[count];
						coordinateIndex[l+2] = polyMesh->GetPolygons()[m].vertices[count+1];
						coordinateIndex[l+3] = -1;
						l+=4;
						count++;
					}
				}
				
				float stiffness;
				float friction;
				float viscosity;	
				bool stiffnessIsSet = false;
				bool frictionIsSet = false;
				bool viscosityIsSet = false;
				
				// Parse the object for diffrent tags and values.
				csRef<iObjectIterator> it = myObject->GetIterator();
				while (!it->IsFinished()){
					
					csRef<iKeyValuePair> kp (SCF_QUERY_INTERFACE (it->GetObject(), iKeyValuePair));
					if (kp)
					{
						
						const char* val = kp->GetValue ();
						const char* key = kp->GetKey ();
						if (!strcmp (key, "stiffness"))
						{
							csScanStr (val, "%f", &stiffness);
							stiffnessIsSet = true;
						}

						if (!strcmp (key, "friction"))
						{
							csScanStr (val, "%f", &friction);
							frictionIsSet = true;
						}
						if (!strcmp (key, "viscosity")) {
							csScanStr (val, "%f", &viscosity);
							viscosityIsSet = true;
						}
					}
					it->Next();
				}
				
				// Parse the name for what kind of object it is.
				etObject* object;
				if (strcmp(myObject->GetName(),"liquid") == 0) { 
					etLiquidTrisurface* obj = new etLiquidTrisurface();
					obj->create(vertexVector,coordinateIndex,true);
					obj->setScale(1);
					if (viscosityIsSet)
						obj->setViscosity(viscosity);
					object = obj;		
				}
				else if(strcmp(myObject->GetName(),"box") == 0) {
					etBox* obj = new etBox();
					csBox3 box;
					meshWrapper->GetWorldBoundingBox(box);
					csVector3 size(box.MaxX()-box.MinX(),box.MaxY()-box.MinY(),box.MaxZ()-box.MinZ());
					obj->setSizeRaw(etVector(size.x, size.y, size.z));	
					object = obj;
				}
				else if(strcmp(myObject->GetName(),"sphere") == 0) {
					etSphere* obj = new etSphere();
					csRef<iBallState> ballstate (SCF_QUERY_INTERFACE(meshWrapper->GetMeshObject(),iBallState));
					float sx;float sy;float sz;
					ballstate->GetRadius(sx,sy,sz);
					obj->setRadiusRaw(sx);
					object = obj;
				}
				else if(strcmp(myObject->GetName(),"cylinder") == 0) {
					etCylinder* obj = new etCylinder();
					csRef<iBallState> ballstate (SCF_QUERY_INTERFACE(meshWrapper->GetMeshObject(),iBallState));
					float sx;float sy;float sz;
					ballstate->GetRadius(sx,sy,sz);
					obj->setRadiusRaw(sx);
					obj->setHeightRaw(sx+sx);
					object = obj;
				}
				else {
					etHapticTrisurface* obj = new etHapticTrisurface();
					obj->create(vertexVector,coordinateIndex,true);
					obj->setScale(1);
					obj->enableBackFaceTouch();
					obj->disableFrontFaceTouch();
					if(stiffnessIsSet) {				
						obj->setRawStiffness(stiffness);
					}
					object = obj;
				}	
				
				// Set default friction if not set.
				if(frictionIsSet) {
					object->setFriction(friction);
				}
								
				csRef<iThingState> thing_state (SCF_QUERY_INTERFACE (meshWrapper->GetMeshObject (),iThingState));
				csRef<iBallState> ball_state (SCF_QUERY_INTERFACE (meshWrapper->GetMeshObject (),iBallState));
				csRef<iSprite3DState> spstate (SCF_QUERY_INTERFACE (meshWrapper->GetMeshObject (), iSprite3DState));
				
				// Add object to the dynamic object list. (Always checked for collision)
				if((thing_state != NULL && thing_state->GetMovingOption() > 0) || spstate != NULL || ball_state!=NULL) { // sprites are moveable and things with moving option > 0
					dynamicObjectList.push_back(object);
					dynamicIndexList.push_back(i);
					nbrOfDynamicObjects++;
				}
				// Add object to the walls object list. (Always checked for collision)
				else if(strcmp(myObject->GetName(),"static") == 0) {
					wallsList.push_back(object);
				}
				// Add object to the static object list. (Octree used for collision detection)
				else {
					csBox3 boundBox; 
					meshWrapper->GetWorldBoundingBox(boundBox);
					boundBoxes.push_back(boundBox);
					staticObjectList.push_back(object);
				}
			}
		}
	}

	// Copy lists to vectors.
	
	std::vector<etObject*> staticObjectVector(staticObjectList.size());
	std::vector<etObject*> dynamicObjectVector(dynamicObjectList.size());
	std::vector<int> dynamicIndexVector(dynamicIndexList.size());
	std::vector<csBox3> objectBoundBoxes(staticObjectList.size());
	std::vector<etObject*> tempWallsVector(wallsList.size());
	
	int objectIndex = 0;
	
	while(!staticObjectList.empty()) {
		staticObjectVector[objectIndex] = staticObjectList.front();
		staticObjectList.pop_front();
		objectIndex++;
	}
	objectIndex = 0;
	
	while(!wallsList.empty()) {
		tempWallsVector[objectIndex] = wallsList.front();
		wallsList.pop_front();
		objectIndex++;
	}
	objectIndex = 0;
	
	while(!dynamicObjectList.empty()) {
		dynamicObjectVector[objectIndex] = dynamicObjectList.front();
		dynamicObjectList.pop_front();
		objectIndex++;
	}
	objectIndex = 0;
	
	while(!boundBoxes.empty()) {
		objectBoundBoxes[objectIndex] = boundBoxes.front();
		boundBoxes.pop_front();
		objectIndex++;
	}
	objectIndex = 0;
	
	while(!dynamicIndexList.empty()) {
		dynamicIndexVector[objectIndex] = dynamicIndexList.front();
		dynamicIndexList.pop_front();
		objectIndex++;
	}
	
	meshList = csMeshList;
	wallsVector = tempWallsVector;
	addStaticObjects(staticObjectVector,objectBoundBoxes,minMaxVec);
	addDynamicObjects(dynamicObjectVector,dynamicIndexVector);
}

/** Returns the collected forces of all objects. */
etVector World::calculateForces(etVector point) {
	
	if (!forcesOn)
		return etVector(0,0,0);
	
	etVector totalForce(0,0,0);
	etVector force(0,0,0);
	csVector3 csPoint(point.x(),point.y(),-point.z());	
	std::list<int> staticIndex;
	std::list<int>::iterator i;
	staticIndex = getNearbyStaticObjects(point.x(),point.y(),-point.z());
	
	// Checking walls
	for (int k = 0; k != wallsVector.size(); k++) {
		force = wallsVector[k]->computePointForce(point);
		
		if(!checkLimit(force))
			force = etVector(0,0,0);
		totalForce += force;
	}
	
	// Checking static objects
	for (i = staticIndex.begin(); i != staticIndex.end(); ++i) {
		force = staticObjects[*i]->computePointForce(point);
		
		if(!checkLimit(force))
			force = etVector(0,0,0);	
		totalForce += force;
	}
	
	mutex->lock(); 
	dynamicForceList.clear();

	// Checking dynamic objects
	for (int j=0;j<nbrOfDynamicObjects;j++) {
		csVector3 pos = meshList->Get(dynamicIndices[j])->GetMovable()->GetTransform().Other2This(csPoint);
		etVector newPoint(pos.x,pos.y,-pos.z); 
		force = dynamicObjects[j]->computePointForce(newPoint);
		csVector3 newForce(force.x(),force.y(),-force.z());
		
		csMatrix3 t2o = meshList->Get(dynamicIndices[j])->GetMovable()->GetTransform().GetT2O();
		newForce = t2o*(t2o,newForce);
		
		etVector trueForce(newForce.x,newForce.y,-newForce.z);
		
		if(!checkLimit(trueForce))
			trueForce = etVector(0,0,0);	
		
		if(trueForce.x() != 0 || trueForce.y() != 0 || trueForce.z() != 0) {
			csVector3 csForce(-trueForce.x()*0.25, -trueForce.y()*0.25, trueForce.z()*0.25);
			std::pair<int,csVector3> forcePair(dynamicIndices[j],csForce);
			dynamicForceList.push_back(forcePair);
		}
		totalForce += trueForce;
	}
	mutex->unlock(); 
	
	if(!checkLimit(totalForce))
		totalForce = etVector(0,0,0);
	
	return totalForce;
}

/** Check if the force is over the max force limit. */
bool World::checkLimit(etVector force) {
	if(force.x() < -MAX_FORCE || force.x() > MAX_FORCE || 
		force.y() < -MAX_FORCE || force.y() > MAX_FORCE|| 
		force.z() < -MAX_FORCE || force.z() > MAX_FORCE) 
	{
		return false;
	}
	return true;
	
}

/** Returns the list containg dynamic objects that have been moved. */
std::pair<int,csVector3> World::getDynamicForcePair() const {
	mutex->lock(); 
	std::pair<int,csVector3> pair = dynamicForcePair;
	mutex->unlock();
	return pair;
}

/** Returns the list containg dynamic objects that have been moved. */
std::list<std::pair<int,csVector3> > World::getDynamicForceList() const {
	mutex->lock(); 
	std::list<std::pair<int,csVector3> > list = dynamicForceList;
	mutex->unlock();
	return list;
}
